import numbers

import numpy

import cupy


###############################################################################
# Private utility functions.


def _round_if_needed(arr, dtype):
    """Rounds arr inplace if the destination dtype is an integer.
    """
    if cupy.issubdtype(dtype, cupy.integer):
        arr.round(out=arr)  # bug in round so use rint (cupy/cupy#2330)


def _slice_at_axis(sl, axis):
    """Constructs a tuple of slices to slice an array in the given dimension.

    Args:
      sl(slice): The slice for the given dimension.
      axis(int): The axis to which `sl` is applied. All other dimensions are
          left "unsliced".

    Returns:
      tuple of slices: A tuple with slices matching `shape` in length.
    """
    return (slice(None),) * axis + (sl,) + (Ellipsis,)


def _view_roi(array, original_area_slice, axis):
    """Gets a view of the current region of interest during iterative padding.

    When padding multiple dimensions iteratively corner values are
    unnecessarily overwritten multiple times. This function reduces the
    working area for the first dimensions so that corners are excluded.

    Args:
      array(cupy.ndarray): The array with the region of interest.
      original_area_slice(tuple of slices): Denotes the area with original
          values of the unpadded array.
      axis(int): The currently padded dimension assuming that `axis` is padded
          before `axis` + 1.

    Returns:
    """
    axis += 1
    sl = (slice(None),) * axis + original_area_slice[axis:]
    return array[sl]


def _pad_simple(array, pad_width, fill_value=None):
    """Pads an array on all sides with either a constant or undefined values.

    Args:
      array(cupy.ndarray): Array to grow.
      pad_width(sequence of tuple[int, int]): Pad width on both sides for each
          dimension in `arr`.
      fill_value(scalar, optional): If provided the padded area is
          filled with this value, otherwise the pad area left undefined.
          (Default value = None)
    """
    # Allocate grown array
    new_shape = tuple(
        left + size + right
        for size, (left, right) in zip(array.shape, pad_width)
    )
    order = 'F' if array.flags.fnc else 'C'  # Fortran and not also C-order
    padded = cupy.empty(new_shape, dtype=array.dtype, order=order)

    if fill_value is not None:
        padded.fill(fill_value)

    # Copy old array into correct space
    original_area_slice = tuple(
        slice(left, left + size)
        for size, (left, right) in zip(array.shape, pad_width)
    )
    padded[original_area_slice] = array

    return padded, original_area_slice


def _set_pad_area(padded, axis, width_pair, value_pair):
    """Set an empty-padded area in given dimension.
    """
    left_slice = _slice_at_axis(slice(None, width_pair[0]), axis)
    padded[left_slice] = value_pair[0]

    right_slice = _slice_at_axis(
        slice(padded.shape[axis] - width_pair[1], None), axis
    )
    padded[right_slice] = value_pair[1]


def _get_edges(padded, axis, width_pair):
    """Retrieves edge values from an empty-padded array along a given axis.

    Args:
      padded(cupy.ndarray): Empty-padded array.
      axis(int): Dimension in which the edges are considered.
      width_pair((int, int)): Pair of widths that mark the pad area on both
          sides in the given dimension.
    """
    left_index = width_pair[0]
    left_slice = _slice_at_axis(slice(left_index, left_index + 1), axis)
    left_edge = padded[left_slice]

    right_index = padded.shape[axis] - width_pair[1]
    right_slice = _slice_at_axis(slice(right_index - 1, right_index), axis)
    right_edge = padded[right_slice]

    return left_edge, right_edge


def _get_linear_ramps(padded, axis, width_pair, end_value_pair):
    """Constructs linear ramps for an empty-padded array along a given axis.

    Args:
      padded(cupy.ndarray): Empty-padded array.
      axis(int): Dimension in which the ramps are constructed.
      width_pair((int, int)): Pair of widths that mark the pad area on both
          sides in the given dimension.
      end_value_pair((scalar, scalar)): End values for the linear ramps which
          form the edge of the fully padded array. These values are included in
          the linear ramps.
    """
    edge_pair = _get_edges(padded, axis, width_pair)

    left_ramp = cupy.linspace(
        start=end_value_pair[0],
        # squeeze axis replaced by linspace
        stop=edge_pair[0].squeeze(axis),
        num=width_pair[0],
        endpoint=False,
        dtype=padded.dtype,
        axis=axis,
    )

    right_ramp = cupy.linspace(
        start=end_value_pair[1],
        # squeeze axis replaced by linspace
        stop=edge_pair[1].squeeze(axis),
        num=width_pair[1],
        endpoint=False,
        dtype=padded.dtype,
        axis=axis,
    )
    # Reverse linear space in appropriate dimension
    right_ramp = right_ramp[_slice_at_axis(slice(None, None, -1), axis)]

    return left_ramp, right_ramp


def _get_stats(padded, axis, width_pair, length_pair, stat_func):
    """Calculates a statistic for an empty-padded array along a given axis.

    Args:
      padded(cupy.ndarray): Empty-padded array.
      axis(int): Dimension in which the statistic is calculated.
      width_pair((int, int)): Pair of widths that mark the pad area on both
          sides in the given dimension.
      length_pair(2-element sequence of None or int): Gives the number of
          values in valid area from each side that is taken into account when
          calculating the statistic. If None the entire valid area in `padded`
          is considered.
      stat_func(function): Function to compute statistic. The expected
          signature is
          ``stat_func(x: ndarray, axis: int, keepdims: bool) -> ndarray``.
    """
    # Calculate indices of the edges of the area with original values
    left_index = width_pair[0]
    right_index = padded.shape[axis] - width_pair[1]
    # as well as its length
    max_length = right_index - left_index

    # Limit stat_lengths to max_length
    left_length, right_length = length_pair
    if left_length is None or max_length < left_length:
        left_length = max_length
    if right_length is None or max_length < right_length:
        right_length = max_length

    # Calculate statistic for the left side
    left_slice = _slice_at_axis(
        slice(left_index, left_index + left_length), axis
    )
    left_chunk = padded[left_slice]
    left_stat = stat_func(left_chunk, axis=axis, keepdims=True)
    _round_if_needed(left_stat, padded.dtype)

    if left_length == right_length == max_length:
        # return early as right_stat must be identical to left_stat
        return left_stat, left_stat

    # Calculate statistic for the right side
    right_slice = _slice_at_axis(
        slice(right_index - right_length, right_index), axis
    )
    right_chunk = padded[right_slice]
    right_stat = stat_func(right_chunk, axis=axis, keepdims=True)
    _round_if_needed(right_stat, padded.dtype)
    return left_stat, right_stat


def _set_reflect_both(padded, axis, width_pair, method, include_edge=False):
    """Pads an `axis` of `arr` using reflection.

    Args:
      padded(cupy.ndarray): Input array of arbitrary shape.
      axis(int): Axis along which to pad `arr`.
      width_pair((int, int)): Pair of widths that mark the pad area on both
          sides in the given dimension.
      method(str): Controls method of reflection; options are 'even' or 'odd'.
      include_edge(bool, optional): If true, edge value is included in
          reflection, otherwise the edge value forms the symmetric axis to the
          reflection. (Default value = False)
    """
    left_pad, right_pad = width_pair
    old_length = padded.shape[axis] - right_pad - left_pad

    if include_edge:
        # Edge is included, we need to offset the pad amount by 1
        edge_offset = 1
    else:
        edge_offset = 0  # Edge is not included, no need to offset pad amount
        old_length -= 1  # but must be omitted from the chunk

    if left_pad > 0:
        # Pad with reflected values on left side:
        # First limit chunk size which can't be larger than pad area
        chunk_length = min(old_length, left_pad)
        # Slice right to left, stop on or next to edge, start relative to stop
        stop = left_pad - edge_offset
        start = stop + chunk_length
        left_slice = _slice_at_axis(slice(start, stop, -1), axis)
        left_chunk = padded[left_slice]

        if method == 'odd':
            # Negate chunk and align with edge
            edge_slice = _slice_at_axis(slice(left_pad, left_pad + 1), axis)
            left_chunk = 2 * padded[edge_slice] - left_chunk

        # Insert chunk into padded area
        start = left_pad - chunk_length
        stop = left_pad
        pad_area = _slice_at_axis(slice(start, stop), axis)
        padded[pad_area] = left_chunk
        # Adjust pointer to left edge for next iteration
        left_pad -= chunk_length

    if right_pad > 0:
        # Pad with reflected values on right side:
        # First limit chunk size which can't be larger than pad area
        chunk_length = min(old_length, right_pad)
        # Slice right to left, start on or next to edge, stop relative to start
        start = -right_pad + edge_offset - 2
        stop = start - chunk_length
        right_slice = _slice_at_axis(slice(start, stop, -1), axis)
        right_chunk = padded[right_slice]

        if method == 'odd':
            # Negate chunk and align with edge
            edge_slice = _slice_at_axis(
                slice(-right_pad - 1, -right_pad), axis
            )
            right_chunk = 2 * padded[edge_slice] - right_chunk

        # Insert chunk into padded area
        start = padded.shape[axis] - right_pad
        stop = start + chunk_length
        pad_area = _slice_at_axis(slice(start, stop), axis)
        padded[pad_area] = right_chunk
        # Adjust pointer to right edge for next iteration
        right_pad -= chunk_length

    return left_pad, right_pad


def _set_wrap_both(padded, axis, width_pair):
    """Pads an `axis` of `arr` with wrapped values.

    Args:
      padded(cupy.ndarray): Input array of arbitrary shape.
      axis(int): Axis along which to pad `arr`.
      width_pair((int, int)): Pair of widths that mark the pad area on both
          sides in the given dimension.
    """
    left_pad, right_pad = width_pair
    period = padded.shape[axis] - right_pad - left_pad

    # If the current dimension of `arr` doesn't contain enough valid values
    # (not part of the undefined pad area) we need to pad multiple times.
    # Each time the pad area shrinks on both sides which is communicated with
    # these variables.
    new_left_pad = 0
    new_right_pad = 0

    if left_pad > 0:
        # Pad with wrapped values on left side
        # First slice chunk from right side of the non-pad area.
        # Use min(period, left_pad) to ensure that chunk is not larger than
        # pad area
        right_slice = _slice_at_axis(
            slice(
                -right_pad - min(period, left_pad),
                -right_pad if right_pad != 0 else None,
            ),
            axis,
        )
        right_chunk = padded[right_slice]

        if left_pad > period:
            # Chunk is smaller than pad area
            pad_area = _slice_at_axis(slice(left_pad - period, left_pad), axis)
            new_left_pad = left_pad - period
        else:
            # Chunk matches pad area
            pad_area = _slice_at_axis(slice(None, left_pad), axis)
        padded[pad_area] = right_chunk

    if right_pad > 0:
        # Pad with wrapped values on right side
        # First slice chunk from left side of the non-pad area.
        # Use min(period, right_pad) to ensure that chunk is not larger than
        # pad area
        left_slice = _slice_at_axis(
            slice(left_pad, left_pad + min(period, right_pad)), axis
        )
        left_chunk = padded[left_slice]

        if right_pad > period:
            # Chunk is smaller than pad area
            pad_area = _slice_at_axis(
                slice(-right_pad, -right_pad + period), axis
            )
            new_right_pad = right_pad - period
        else:
            # Chunk matches pad area
            pad_area = _slice_at_axis(slice(-right_pad, None), axis)
        padded[pad_area] = left_chunk

    return new_left_pad, new_right_pad


def _as_pairs(x, ndim, as_index=False):
    """Broadcasts `x` to an array with shape (`ndim`, 2).

    A helper function for `pad` that prepares and validates arguments like
    `pad_width` for iteration in pairs.

    Args:
      x(scalar or array-like, optional): The object to broadcast to the shape
          (`ndim`, 2).
      ndim(int): Number of pairs the broadcasted `x` will have.
      as_index(bool, optional): If `x` is not None, try to round each
          element of `x` to an integer (dtype `cupy.intp`) and ensure every
          element is positive. (Default value = False)

    Returns:
      nested iterables, shape (`ndim`, 2): The broadcasted version of `x`.
    """
    if x is None:
        # Pass through None as a special case, otherwise cupy.round(x) fails
        # with an AttributeError
        return ((None, None),) * ndim
    elif isinstance(x, numbers.Number):
        if as_index:
            x = round(x)
        return ((x, x),) * ndim

    x = numpy.array(x)
    if as_index:
        x = numpy.asarray(numpy.round(x), dtype=numpy.intp)

    if x.ndim < 3:
        # Optimization: Possibly use faster paths for cases where `x` has
        # only 1 or 2 elements. `numpy.broadcast_to` could handle these as well
        # but is currently slower

        if x.size == 1:
            # x was supplied as a single value
            x = x.ravel()  # Ensure x[0] works for x.ndim == 0, 1, 2
            if as_index and x < 0:
                raise ValueError("index can't contain negative values")
            return ((x[0], x[0]),) * ndim

        if x.size == 2 and x.shape != (2, 1):
            # x was supplied with a single value for each side
            # but except case when each dimension has a single value
            # which should be broadcasted to a pair,
            # e.g. [[1], [2]] -> [[1, 1], [2, 2]] not [[1, 2], [1, 2]]
            x = x.ravel()  # Ensure x[0], x[1] works
            if as_index and (x[0] < 0 or x[1] < 0):
                raise ValueError("index can't contain negative values")
            return ((x[0], x[1]),) * ndim

    if as_index and x.min() < 0:
        raise ValueError("index can't contain negative values")

    # Converting the array with `tolist` seems to improve performance
    # when iterating and indexing the result (see usage in `pad`)
    x_view = x.view()
    x_view.shape = (ndim, 2)
    return x_view.tolist()

# def _pad_dispatcher(array, pad_width, mode=None, **kwargs):
#    return (array,)


###############################################################################
# Public functions


# @array_function_dispatch(_pad_dispatcher, module='numpy')
def pad(array, pad_width, mode='constant', **kwargs):
    """Pads an array with specified widths and values.

    Args:
      array(cupy.ndarray): The array to pad.
      pad_width(sequence, array_like or int): Number of values padded to the
          edges of each axis. ((before_1, after_1), ... (before_N, after_N))
          unique pad widths for each axis. ((before, after),) yields same
          before and after pad for each axis. (pad,) or int is a shortcut for
          before = after = pad width for all axes. You cannot specify
          ``cupy.ndarray``.
      mode(str or function, optional): One of the following string values or a
          user supplied function

          'constant' (default)
              Pads with a constant value.
          'edge'
              Pads with the edge values of array.
          'linear_ramp'
              Pads with the linear ramp between end_value and the array edge
              value.
          'maximum'
              Pads with the maximum value of all or part of the vector along
              each axis.
          'mean'
              Pads with the mean value of all or part of the vector along each
              axis.
          'median'
              Pads with the median value of all or part of the vector along
              each axis. (Not Implemented)
          'minimum'
              Pads with the minimum value of all or part of the vector along
              each axis.
          'reflect'
              Pads with the reflection of the vector mirrored on the first and
              last values of the vector along each axis.
          'symmetric'
               Pads with the reflection of the vector mirrored along the edge
               of the array.
          'wrap'
              Pads with the wrap of the vector along the axis. The first
              values are used to pad the end and the end values are used to
              pad the beginning.
          'empty'
              Pads with undefined values.
          <function>
              Padding function, see Notes.
      stat_length(sequence or int, optional): Used in 'maximum', 'mean',
          'median', and 'minimum'.  Number of values at edge of each axis used
          to calculate the statistic value.
          ((before_1, after_1), ... (before_N, after_N)) unique statistic
          lengths for each axis. ((before, after),) yields same before and
          after statistic lengths for each axis. (stat_length,) or int is a
          shortcut for before = after = statistic length for all axes.
          Default is ``None``, to use the entire axis. You cannot specify
          ``cupy.ndarray``.
      constant_values(sequence or scalar, optional): Used in 'constant'. The
          values to set the padded values for each axis.
          ((before_1, after_1), ... (before_N, after_N)) unique pad constants
          for each axis.
          ((before, after),) yields same before and after constants for each
          axis.
          (constant,) or constant is a shortcut for before = after = constant
          for all axes.
          Default is 0. You cannot specify ``cupy.ndarray``.
      end_values(sequence or scalar, optional): Used in 'linear_ramp'. The
          values used for the ending value of the linear_ramp and that will
          form the edge of the padded array.
          ((before_1, after_1), ... (before_N, after_N)) unique end values
          for each axis.
          ((before, after),) yields same before and after end
          values for each axis.
          (constant,) or constant is a shortcut for before = after = constant
          for all axes.
          Default is 0. You cannot specify ``cupy.ndarray``.
      reflect_type({'even', 'odd'}, optional): Used in 'reflect', and
          'symmetric'.  The 'even' style is the default with an unaltered
          reflection around the edge value.  For the 'odd' style, the extended
          part of the array is created by subtracting the reflected values from
          two times the edge value.

    Returns:
      cupy.ndarray: Padded array with shape extended by ``pad_width``.

    .. note::
        For an array with rank greater than 1, some of the padding of later
        axes is calculated from padding of previous axes.  This is easiest to
        think about with a rank 2 array where the corners of the padded array
        are calculated by using padded values from the first axis.

        The padding function, if used, should modify a rank 1 array in-place.
        It has the following signature:

        ``padding_func(vector, iaxis_pad_width, iaxis, kwargs)``

        where

        vector (cupy.ndarray)
            A rank 1 array already padded with zeros.  Padded values are
            ``vector[:iaxis_pad_width[0]]`` and
            ``vector[-iaxis_pad_width[1]:]``.
        iaxis_pad_width (tuple)
            A 2-tuple of ints, ``iaxis_pad_width[0]`` represents the number of
            values padded at the beginning of vector where
            ``iaxis_pad_width[1]`` represents the number of values padded at
            the end of vector.
        iaxis (int)
            The axis currently being calculated.
        kwargs (dict)
            Any keyword arguments the function requires.

    Examples
    --------
    >>> a = cupy.array([1, 2, 3, 4, 5])
    >>> cupy.pad(a, (2, 3), 'constant', constant_values=(4, 6))
    array([4, 4, 1, ..., 6, 6, 6])

    >>> cupy.pad(a, (2, 3), 'edge')
    array([1, 1, 1, ..., 5, 5, 5])

    >>> cupy.pad(a, (2, 3), 'linear_ramp', end_values=(5, -4))
    array([ 5,  3,  1,  2,  3,  4,  5,  2, -1, -4])

    >>> cupy.pad(a, (2,), 'maximum')
    array([5, 5, 1, 2, 3, 4, 5, 5, 5])

    >>> cupy.pad(a, (2,), 'mean')
    array([3, 3, 1, 2, 3, 4, 5, 3, 3])

    >>> a = cupy.array([[1, 2], [3, 4]])
    >>> cupy.pad(a, ((3, 2), (2, 3)), 'minimum')
    array([[1, 1, 1, 2, 1, 1, 1],
           [1, 1, 1, 2, 1, 1, 1],
           [1, 1, 1, 2, 1, 1, 1],
           [1, 1, 1, 2, 1, 1, 1],
           [3, 3, 3, 4, 3, 3, 3],
           [1, 1, 1, 2, 1, 1, 1],
           [1, 1, 1, 2, 1, 1, 1]])

    >>> a = cupy.array([1, 2, 3, 4, 5])
    >>> cupy.pad(a, (2, 3), 'reflect')
    array([3, 2, 1, 2, 3, 4, 5, 4, 3, 2])

    >>> cupy.pad(a, (2, 3), 'reflect', reflect_type='odd')
    array([-1,  0,  1,  2,  3,  4,  5,  6,  7,  8])

    >>> cupy.pad(a, (2, 3), 'symmetric')
    array([2, 1, 1, 2, 3, 4, 5, 5, 4, 3])

    >>> cupy.pad(a, (2, 3), 'symmetric', reflect_type='odd')
    array([0, 1, 1, 2, 3, 4, 5, 5, 6, 7])

    >>> cupy.pad(a, (2, 3), 'wrap')
    array([4, 5, 1, 2, 3, 4, 5, 1, 2, 3])

    >>> def pad_with(vector, pad_width, iaxis, kwargs):
    ...     pad_value = kwargs.get('padder', 10)
    ...     vector[:pad_width[0]] = pad_value
    ...     vector[-pad_width[1]:] = pad_value
    >>> a = cupy.arange(6)
    >>> a = a.reshape((2, 3))
    >>> cupy.pad(a, 2, pad_with)
    array([[10, 10, 10, 10, 10, 10, 10],
           [10, 10, 10, 10, 10, 10, 10],
           [10, 10,  0,  1,  2, 10, 10],
           [10, 10,  3,  4,  5, 10, 10],
           [10, 10, 10, 10, 10, 10, 10],
           [10, 10, 10, 10, 10, 10, 10]])
    >>> cupy.pad(a, 2, pad_with, padder=100)
    array([[100, 100, 100, 100, 100, 100, 100],
           [100, 100, 100, 100, 100, 100, 100],
           [100, 100,   0,   1,   2, 100, 100],
           [100, 100,   3,   4,   5, 100, 100],
           [100, 100, 100, 100, 100, 100, 100],
           [100, 100, 100, 100, 100, 100, 100]])
    """
    if isinstance(pad_width, numbers.Integral):
        pad_width = ((pad_width, pad_width),) * array.ndim
    else:
        pad_width = numpy.asarray(pad_width)

        if not pad_width.dtype.kind == 'i':
            raise TypeError('`pad_width` must be of integral type.')

        # Broadcast to shape (array.ndim, 2)
        pad_width = _as_pairs(pad_width, array.ndim, as_index=True)

    if callable(mode):
        # Old behavior: Use user-supplied function with numpy.apply_along_axis
        function = mode
        # Create a new zero padded array
        padded, _ = _pad_simple(array, pad_width, fill_value=0)
        # And apply along each axis

        for axis in range(padded.ndim):
            # Iterate using ndindex as in apply_along_axis, but assuming that
            # function operates inplace on the padded array.

            # view with the iteration axis at the end
            view = cupy.moveaxis(padded, axis, -1)

            # compute indices for the iteration axes, and append a trailing
            # ellipsis to prevent 0d arrays decaying to scalars (gh-8642)
            inds = numpy.ndindex(view.shape[:-1])
            inds = (ind + (Ellipsis,) for ind in inds)
            for ind in inds:
                function(view[ind], pad_width[axis], axis, kwargs)

        return padded

    # Make sure that no unsupported keywords were passed for the current mode
    allowed_kwargs = {
        'empty': [],
        'edge': [],
        'wrap': [],
        'constant': ['constant_values'],
        'linear_ramp': ['end_values'],
        'maximum': ['stat_length'],
        'mean': ['stat_length'],
        # 'median': ['stat_length'],
        'minimum': ['stat_length'],
        'reflect': ['reflect_type'],
        'symmetric': ['reflect_type'],
    }
    try:
        unsupported_kwargs = set(kwargs) - set(allowed_kwargs[mode])
    except KeyError:
        raise ValueError("mode '{}' is not supported".format(mode))
    if unsupported_kwargs:
        raise ValueError(
            "unsupported keyword arguments for mode '{}': {}".format(
                mode, unsupported_kwargs
            )
        )

    if mode == 'constant':
        values = kwargs.get('constant_values', 0)
        if isinstance(values, numbers.Number) and values == 0 and (
                array.ndim == 1 or array.size < 4e6):
            # faster path for 1d arrays or small n-dimensional arrays
            return _pad_simple(array, pad_width, 0)[0]

    stat_functions = {
        'maximum': cupy.max,
        'minimum': cupy.min,
        'mean': cupy.mean,
        # 'median': cupy.median,
    }

    # Create array with final shape and original values
    # (padded area is undefined)
    padded, original_area_slice = _pad_simple(array, pad_width)
    # And prepare iteration over all dimensions
    # (zipping may be more readable than using enumerate)
    axes = range(padded.ndim)

    if mode == 'constant':
        values = _as_pairs(values, padded.ndim)
        for axis, width_pair, value_pair in zip(axes, pad_width, values):
            roi = _view_roi(padded, original_area_slice, axis)
            _set_pad_area(roi, axis, width_pair, value_pair)

    elif mode == 'empty':
        pass  # Do nothing as _pad_simple already returned the correct result

    elif array.size == 0:
        # Only modes 'constant' and 'empty' can extend empty axes, all other
        # modes depend on `array` not being empty
        # -> ensure every empty axis is only 'padded with 0'
        for axis, width_pair in zip(axes, pad_width):
            if array.shape[axis] == 0 and any(width_pair):
                raise ValueError(
                    "can't extend empty axis {} using modes other than "
                    "'constant' or 'empty'".format(axis)
                )
        # passed, don't need to do anything more as _pad_simple already
        # returned the correct result

    elif mode == 'edge':
        for axis, width_pair in zip(axes, pad_width):
            roi = _view_roi(padded, original_area_slice, axis)
            edge_pair = _get_edges(roi, axis, width_pair)
            _set_pad_area(roi, axis, width_pair, edge_pair)

    elif mode == 'linear_ramp':
        end_values = kwargs.get('end_values', 0)
        end_values = _as_pairs(end_values, padded.ndim)
        for axis, width_pair, value_pair in zip(axes, pad_width, end_values):
            roi = _view_roi(padded, original_area_slice, axis)
            ramp_pair = _get_linear_ramps(roi, axis, width_pair, value_pair)
            _set_pad_area(roi, axis, width_pair, ramp_pair)

    elif mode in stat_functions:
        func = stat_functions[mode]
        length = kwargs.get('stat_length', None)
        length = _as_pairs(length, padded.ndim, as_index=True)
        for axis, width_pair, length_pair in zip(axes, pad_width, length):
            roi = _view_roi(padded, original_area_slice, axis)
            stat_pair = _get_stats(roi, axis, width_pair, length_pair, func)
            _set_pad_area(roi, axis, width_pair, stat_pair)

    elif mode in {'reflect', 'symmetric'}:
        method = kwargs.get('reflect_type', 'even')
        include_edge = True if mode == 'symmetric' else False
        for axis, (left_index, right_index) in zip(axes, pad_width):
            if array.shape[axis] == 1 and (left_index > 0 or right_index > 0):
                # Extending singleton dimension for 'reflect' is legacy
                # behavior; it really should raise an error.
                edge_pair = _get_edges(padded, axis, (left_index, right_index))
                _set_pad_area(
                    padded, axis, (left_index, right_index), edge_pair
                )
                continue

            roi = _view_roi(padded, original_area_slice, axis)
            while left_index > 0 or right_index > 0:
                # Iteratively pad until dimension is filled with reflected
                # values. This is necessary if the pad area is larger than
                # the length of the original values in the current dimension.
                left_index, right_index = _set_reflect_both(
                    roi, axis, (left_index, right_index), method, include_edge
                )

    elif mode == 'wrap':
        for axis, (left_index, right_index) in zip(axes, pad_width):
            roi = _view_roi(padded, original_area_slice, axis)
            while left_index > 0 or right_index > 0:
                # Iteratively pad until dimension is filled with wrapped
                # values. This is necessary if the pad area is larger than
                # the length of the original values in the current dimension.
                left_index, right_index = _set_wrap_both(
                    roi, axis, (left_index, right_index)
                )

    return padded



import unittest

import numpy
import pytest

import cupy
from cupy import testing


@testing.gpu
class TestTranspose(unittest.TestCase):

    @testing.numpy_cupy_array_equal()
    def test_moveaxis1(self, xp):
        a = testing.shaped_arange((2, 3, 4), xp)
        return xp.moveaxis(a, [0, 1], [1, 2])

    @testing.numpy_cupy_array_equal()
    def test_moveaxis2(self, xp):
        a = testing.shaped_arange((2, 3, 4), xp)
        return xp.moveaxis(a, 1, -1)

    @testing.numpy_cupy_array_equal()
    def test_moveaxis3(self, xp):
        a = testing.shaped_arange((2, 3, 4), xp)
        return xp.moveaxis(a, [0, 2], [1, 0])

    @testing.numpy_cupy_array_equal()
    def test_moveaxis4(self, xp):
        a = testing.shaped_arange((2, 3, 4), xp)
        return xp.moveaxis(a, [2, 0], [1, 0])

    @testing.numpy_cupy_array_equal()
    def test_moveaxis5(self, xp):
        a = testing.shaped_arange((2, 3, 4), xp)
        return xp.moveaxis(a, [2, 0], [0, 1])

    @testing.numpy_cupy_array_equal()
    def test_moveaxis6(self, xp):
        a = testing.shaped_arange((2, 3, 4, 5, 6), xp)
        return xp.moveaxis(a, [0, 2, 1], [3, 4, 0])

    # dim is too large
    def test_moveaxis_invalid1_1(self):
        for xp in (numpy, cupy):
            a = testing.shaped_arange((2, 3, 4), xp)
            with pytest.raises(numpy.AxisError):
                xp.moveaxis(a, [0, 1], [1, 3])

    def test_moveaxis_invalid1_2(self):
        for xp in (numpy, cupy):
            a = testing.shaped_arange((2, 3, 4), xp)
            with pytest.raises(numpy.AxisError):
                xp.moveaxis(a, [0, 1], [1, 3])

    def test_moveaxis_invalid1_3(self):
        for xp in (numpy, cupy):
            a = testing.shaped_arange((2, 3, 4), xp)
            with pytest.raises(numpy.AxisError):
                xp.moveaxis(a, 0, 3)

    # dim is too small
    def test_moveaxis_invalid2_1(self):
        for xp in (numpy, cupy):
            a = testing.shaped_arange((2, 3, 4), xp)
            with pytest.raises(numpy.AxisError):
                xp.moveaxis(a, [0, -4], [1, 2])

    def test_moveaxis_invalid2_2(self):
        for xp in (numpy, cupy):
            a = testing.shaped_arange((2, 3, 4), xp)
            with pytest.raises(numpy.AxisError):
                xp.moveaxis(a, [0, -4], [1, 2])

    def test_moveaxis_invalid2_3(self):
        for xp in (numpy, cupy):
            a = testing.shaped_arange((2, 3, 4), xp)
            with pytest.raises(numpy.AxisError):
                xp.moveaxis(a, -4, 0)

    # len(source) != len(destination)
    def test_moveaxis_invalid3_1(self):
        for xp in (numpy, cupy):
            a = testing.shaped_arange((2, 3, 4), xp)
            with pytest.raises(ValueError):
                xp.moveaxis(a, [0, 1, 2], [1, 2])

    def test_moveaxis_invalid3_2(self):
        for xp in (numpy, cupy):
            a = testing.shaped_arange((2, 3, 4), xp)
            with pytest.raises(ValueError):
                xp.moveaxis(a, 0, [1, 2])

    # len(source) != len(destination)
    def test_moveaxis_invalid4_1(self):
        for xp in (numpy, cupy):
            a = testing.shaped_arange((2, 3, 4), xp)
            with pytest.raises(ValueError):
                xp.moveaxis(a, [0, 1], [1, 2, 0])

    def test_moveaxis_invalid4_2(self):
        for xp in (numpy, cupy):
            a = testing.shaped_arange((2, 3, 4), xp)
            with pytest.raises(ValueError):
                xp.moveaxis(a, [0, 1], 1)

    # Use the same axis twice
    def test_moveaxis_invalid5_1(self):
        for xp in (numpy, cupy):
            a = testing.shaped_arange((2, 3, 4), xp)
            with pytest.raises(numpy.AxisError):
                xp.moveaxis(a, [1, -1], [1, 3])

    def test_moveaxis_invalid5_2(self):
        for xp in (numpy, cupy):
            a = testing.shaped_arange((2, 3, 4), xp)
            with pytest.raises(ValueError):
                xp.moveaxis(a, [0, 1], [-1, 2])

    def test_moveaxis_invalid5_3(self):
        for xp in (numpy, cupy):
            a = testing.shaped_arange((2, 3, 4), xp)
            with pytest.raises(ValueError):
                xp.moveaxis(a, [0, 1], [1, 1])

    @testing.numpy_cupy_array_equal()
    def test_rollaxis(self, xp):
        a = testing.shaped_arange((2, 3, 4), xp)
        return xp.rollaxis(a, 2)

    def test_rollaxis_failure(self):
        for xp in (numpy, cupy):
            a = testing.shaped_arange((2, 3, 4), xp)
            with pytest.raises(ValueError):
                xp.rollaxis(a, 3)

    @testing.numpy_cupy_array_equal()
    def test_swapaxes(self, xp):
        a = testing.shaped_arange((2, 3, 4), xp)
        return xp.swapaxes(a, 2, 0)

    def test_swapaxes_failure(self):
        for xp in (numpy, cupy):
            a = testing.shaped_arange((2, 3, 4), xp)
            with pytest.raises(ValueError):
                xp.swapaxes(a, 3, 0)

    @testing.numpy_cupy_array_equal()
    def test_transpose(self, xp):
        a = testing.shaped_arange((2, 3, 4), xp)
        return a.transpose(-1, 0, 1)

    @testing.numpy_cupy_array_equal()
    def test_transpose_empty(self, xp):
        a = testing.shaped_arange((2, 3, 4), xp)
        return a.transpose()

    @testing.numpy_cupy_array_equal()
    def test_transpose_none(self, xp):
        a = testing.shaped_arange((2, 3, 4), xp)
        return a.transpose(None)

    @testing.numpy_cupy_array_equal()
    def test_external_transpose(self, xp):
        a = testing.shaped_arange((2, 3, 4), xp)
        return xp.transpose(a, (-1, 0, 1))

    @testing.numpy_cupy_array_equal()
    def test_external_transpose_all(self, xp):
        a = testing.shaped_arange((2, 3, 4), xp)
        return xp.transpose(a)



import numpy

import cupy
from cupy.cuda import cublas
from cupy.cuda import device
from cupy.linalg import _util


def solve_triangular(a, b, trans=0, lower=False, unit_diagonal=False,
                     overwrite_b=False, check_finite=False):
    """Solve the equation a x = b for x, assuming a is a triangular matrix.

    Args:
        a (cupy.ndarray): The matrix with dimension ``(M, M)``.
        b (cupy.ndarray): The matrix with dimension ``(M,)`` or
            ``(M, N)``.
        lower (bool): Use only data contained in the lower triangle of ``a``.
            Default is to use upper triangle.
        trans (0, 1, 2, 'N', 'T' or 'C'): Type of system to solve:

            - *'0'* or *'N'* -- :math:`a x  = b`
            - *'1'* or *'T'* -- :math:`a^T x = b`
            - *'2'* or *'C'* -- :math:`a^H x = b`

        unit_diagonal (bool): If ``True``, diagonal elements of ``a`` are
            assumed to be 1 and will not be referenced.
        overwrite_b (bool): Allow overwriting data in b (may enhance
            performance)
        check_finite (bool): Whether to check that the input matrices contain
            only finite numbers. Disabling may give a performance gain, but may
            result in problems (crashes, non-termination) if the inputs do
            contain infinities or NaNs.

    Returns:
        cupy.ndarray:
            The matrix with dimension ``(M,)`` or ``(M, N)``.

    .. seealso:: :func:`scipy.linalg.solve_triangular`
    """

    _util._assert_cupy_array(a, b)

    if len(a.shape) != 2 or a.shape[0] != a.shape[1]:
        raise ValueError('expected square matrix')
    if len(a) != len(b):
        raise ValueError('incompatible dimensions')

    # Cast to float32 or float64
    if a.dtype.char in 'fd':
        dtype = a.dtype
    else:
        dtype = numpy.promote_types(a.dtype.char, 'f')

    a = cupy.array(a, dtype=dtype, order='F', copy=False)
    b = cupy.array(b, dtype=dtype, order='F', copy=(not overwrite_b))

    if check_finite:
        if a.dtype.kind == 'f' and not cupy.isfinite(a).all():
            raise ValueError(
                'array must not contain infs or NaNs')
        if b.dtype.kind == 'f' and not cupy.isfinite(b).all():
            raise ValueError(
                'array must not contain infs or NaNs')

    m, n = (b.size, 1) if b.ndim == 1 else b.shape
    cublas_handle = device.get_cublas_handle()

    if dtype == 'f':
        trsm = cublas.strsm
    elif dtype == 'd':
        trsm = cublas.dtrsm
    elif dtype == 'F':
        trsm = cublas.ctrsm
    else:  # dtype == 'D'
        trsm = cublas.ztrsm
    one = numpy.array(1, dtype=dtype)

    if lower:
        uplo = cublas.CUBLAS_FILL_MODE_LOWER
    else:
        uplo = cublas.CUBLAS_FILL_MODE_UPPER

    if trans == 'N':
        trans = cublas.CUBLAS_OP_N
    elif trans == 'T':
        trans = cublas.CUBLAS_OP_T
    elif trans == 'C':
        trans = cublas.CUBLAS_OP_C

    if unit_diagonal:
        diag = cublas.CUBLAS_DIAG_UNIT
    else:
        diag = cublas.CUBLAS_DIAG_NON_UNIT

    trsm(
        cublas_handle, cublas.CUBLAS_SIDE_LEFT, uplo,
        trans, diag,
        m, n, one.ctypes.data, a.data.ptr, m, b.data.ptr, m)
    return b



import unittest

from cupy import testing
import cupyx.scipy.linalg  # NOQA

try:
    import scipy.linalg  # NOQA
except ImportError:
    pass


class TestSpecialMatricesBase(unittest.TestCase):
    def _get_arg(self, xp, arg):
        if isinstance(arg, tuple):
            # Allocate array with the given shape
            return testing.shaped_random(arg, xp)

        # Otherwise just pass the arg back
        return arg


@testing.parameterize(*(
    testing.product({
        # 1 argument: 1D array
        'function': ['circulant', 'toeplitz', 'hankel', 'companion'],
        'args': [((1,),), ((2,),), ((4,),), ((10,),), ((25,),)],
    }) + testing.product({
        # 2 arguments: both 1D arrays
        # For leslie, second array should be 1 less long than first
        'function': ['toeplitz', 'hankel', 'leslie'],
        'args': [((1,), (1,)), ((2,), (1,)), ((4,), (5,)),
                 ((10,), (9,)), ((25,), (24,))],
    }) + testing.product({
        # 1 argument: int
        'function': ['tri', 'hadamard', 'helmert', 'hilbert', 'dft'],
        'args': [(1,), (2,), (4,), (10,), (25,)],
    }) + testing.product({
        # 2 arguments: int, dtype
        # NOTE: for tri this is an undocumented, legacy, call
        'function': ['tri', 'hadamard'],
        'args': [(4, 'int32'), (8, 'float64'), (6, 'float64')],
    }) + testing.product({
        # 2 arguments: int, bool
        'function': ['helmert'],
        'args': [(4, True), (5, False)],
    }) + testing.product({
        # 2 arguments: int, str
        'function': ['dft'],
        'args': [(4, 'sqrtn'), (5, 'n')],
    }) + testing.product({
        # 2-4 arguments: int, int, optional int, optional dtype
        'function': ['tri'],
        'args': [(4, 5), (5, 5), (5, 4),
                 (4, 5, -1), (4, 5, 1), (4, 5, 0), (5, 4, -1), (5, 5, 1),
                 (4, 5, -1, int), (5, 4, 0, float)],
    }) + testing.product({
        # 1-2 arguments: 2D array, optional int
        'function': ['tril', 'triu'],
        'args': [((5, 5),), ((4, 5),), ((5, 4),), ((4, 4),),
                 ((5, 5), -1), ((4, 5), 1), ((5, 4), -1), ((4, 4), 1)],
    }) + testing.product({
        # 2 arguments: both 2D arrays
        'function': ['kron'],
        'args': [((5, 5), (4, 5),), ((5, 4), (4, 4),), ((1, 2), (4, 5))],
    }) + testing.product({
        # 0 or more arguments: all 1 or 2D arrays
        'function': ['block_diag'],
        'args': [(), ((5,),), ((4, 5),), ((5,), (4, 4),), ((1, 2), (4, 5)),
                 ((1,), (4, 5), (3, 6), (7, 2), (8, 9))],
    })
))
@testing.gpu
@testing.with_requires('scipy')
class TestSpecialMatrices(TestSpecialMatricesBase):
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name='scp',
                                 accept_error=ValueError)
    def test_special_matrix(self, xp, scp):
        function = getattr(scp.linalg, self.function)
        return function(*[self._get_arg(xp, arg) for arg in self.args])


@testing.parameterize(*(
    testing.product({
        # 1 argument: 1D array
        'function': ['fiedler', 'fiedler_companion'],
        'args': [((0,),), ((1,),), ((2,),), ((4,),), ((10,),), ((25,),)],
    })
))
@testing.gpu
@testing.with_requires('scipy>=1.3.0')
class TestSpecialMatrices_1_3_0(TestSpecialMatricesBase):
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name='scp',
                                 accept_error=ValueError)
    def test_special_matrix(self, xp, scp):
        function = getattr(scp.linalg, self.function)
        return function(*[self._get_arg(xp, arg) for arg in self.args])

    def _get_arg(self, xp, arg):
        if isinstance(arg, tuple):
            # Allocate array with the given shape
            return testing.shaped_random(arg, xp)

        # Otherwise just pass the arg back
        return arg


@testing.parameterize(*(
    testing.product({
        # 2-3 arguments: 1D array, 1 int, 1 optional str
        'function': ['convolution_matrix'],
        'args': [((1,), 5), ((2,), 3), ((4,), 10), ((10,), 15), ((25,), 25),
                 ((4,), 6, 'full'), ((10,), 8, 'same'), ((25,), 25, 'valid')],
    })
))
@testing.gpu
@testing.with_requires('scipy>=1.5.0')
class TestSpecialMatrices_1_5_0(TestSpecialMatricesBase):
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name='scp',
                                 accept_error=ValueError)
    def test_special_matrix(self, xp, scp):
        function = getattr(scp.linalg, self.function)
        return function(*[self._get_arg(xp, arg) for arg in self.args])



"""
This file must not depend on any other CuPy modules.
"""

import ctypes
import json
import os
import os.path
import shutil
import sys
import warnings


# '' for uninitialized, None for non-existing
_cuda_path = ''
_nvcc_path = ''
_rocm_path = ''
_hipcc_path = ''
_cub_path = ''

"""
Library Preloading
------------------

Wheel packages are built against specific versions of CUDA libraries
(cuTENSOR/NCCL/cuDNN).
To avoid loading wrong version, these shared libraries are manually
preloaded.

# TODO(kmaehashi): Support NCCL

Example of `_preload_config` is as follows:

{
    # installation source
    'packaging': 'pip',

    # CUDA version string
    'cuda': '11.0',

    'cudnn': {
        # cuDNN version string
        'version': '8.0.0',

        # names of the shared library
        'filenames': ['libcudnn.so.X.Y.Z']  # or `cudnn64_X.dll` for Windows
    }
}

The configuration file is intended solely for internal purposes and
not expected to be parsed by end-users.
"""

_preload_config = None

_preload_libs = {
    'cudnn': None,
    'nccl': None,
    'cutensor': None,
}

_preload_logs = []


def _log(msg):
    # TODO(kmaehashi): replace with the standard logging
    _preload_logs.append(msg)


def get_cuda_path():
    # Returns the CUDA installation path or None if not found.
    global _cuda_path
    if _cuda_path == '':
        _cuda_path = _get_cuda_path()
    return _cuda_path


def get_nvcc_path():
    # Returns the path to the nvcc command or None if not found.
    global _nvcc_path
    if _nvcc_path == '':
        _nvcc_path = _get_nvcc_path()
    return _nvcc_path


def get_rocm_path():
    # Returns the ROCm installation path or None if not found.
    global _rocm_path
    if _rocm_path == '':
        _rocm_path = _get_rocm_path()
    return _rocm_path


def get_hipcc_path():
    # Returns the path to the hipcc command or None if not found.
    global _hipcc_path
    if _hipcc_path == '':
        _hipcc_path = _get_hipcc_path()
    return _hipcc_path


def get_cub_path():
    # Returns the CUB header path or None if not found.
    global _cub_path
    if _cub_path == '':
        _cub_path = _get_cub_path()
    return _cub_path


def _get_cuda_path():
    # Use environment variable
    cuda_path = os.environ.get('CUDA_PATH', '')  # Nvidia default on Windows
    if os.path.exists(cuda_path):
        return cuda_path

    # Use nvcc path
    nvcc_path = shutil.which('nvcc')
    if nvcc_path is not None:
        return os.path.dirname(os.path.dirname(nvcc_path))

    # Use typical path
    if os.path.exists('/usr/local/cuda'):
        return '/usr/local/cuda'

    return None


def _get_nvcc_path():
    # Honor the "NVCC" env var
    nvcc_path = os.environ.get('NVCC', None)
    if nvcc_path is not None:
        return nvcc_path

    # Lookup <CUDA>/bin
    cuda_path = get_cuda_path()
    if cuda_path is None:
        return None

    return shutil.which('nvcc', path=os.path.join(cuda_path, 'bin'))


def _get_rocm_path():
    # Use environment variable
    rocm_path = os.environ.get('ROCM_HOME', '')
    if os.path.exists(rocm_path):
        return rocm_path

    # Use hipcc path
    hipcc_path = shutil.which('hipcc')
    if hipcc_path is not None:
        return os.path.dirname(os.path.dirname(hipcc_path))

    # Use typical path
    if os.path.exists('/opt/rocm'):
        return '/opt/rocm'

    return None


def _get_hipcc_path():
    # TODO(leofang): Introduce an env var HIPCC?

    # Lookup <ROCM>/bin
    rocm_path = get_rocm_path()
    if rocm_path is None:
        return None

    return shutil.which('hipcc', path=os.path.join(rocm_path, 'bin'))


def _get_cub_path():
    # runtime discovery of CUB headers
    from cupy_backends.cuda.api import runtime
    current_dir = os.path.dirname(os.path.abspath(__file__))

    if not runtime.is_hip:
        cuda_path = get_cuda_path()
        if os.path.isdir(os.path.join(current_dir, '_core/include/cupy/cub')):
            _cub_path = '<bundle>'
        elif cuda_path is not None and os.path.isdir(
                os.path.join(cuda_path, 'include/cub')):
            # use built-in CUB for CUDA 11+
            _cub_path = '<CUDA>'
        else:
            _cub_path = None
    else:
        # the bundled CUB does not work in ROCm
        rocm_path = get_rocm_path()
        if rocm_path is not None and os.path.isdir(
                os.path.join(rocm_path, 'include/hipcub')):
            # use hipCUB
            _cub_path = '<ROCm>'
        else:
            _cub_path = None
    return _cub_path


def _setup_win32_dll_directory():
    # Setup DLL directory to load CUDA Toolkit libs and shared libraries
    # added during the build process.
    if sys.platform.startswith('win32'):
        is_conda = ((os.environ.get('CONDA_PREFIX') is not None)
                    or (os.environ.get('CONDA_BUILD_STATE') is not None))
        # Path to the CUDA Toolkit binaries
        cuda_path = get_cuda_path()
        if cuda_path is not None:
            if is_conda:
                cuda_bin_path = cuda_path
            else:
                cuda_bin_path = os.path.join(cuda_path, 'bin')
        else:
            cuda_bin_path = None
            warnings.warn(
                'CUDA path could not be detected.'
                ' Set CUDA_PATH environment variable if CuPy fails to load.')
        _log('CUDA_PATH: {}'.format(cuda_path))

        # Path to shared libraries in wheel
        wheel_libdir = os.path.join(
            get_cupy_install_path(), 'cupy', '.data', 'lib')
        if os.path.isdir(wheel_libdir):
            _log('Wheel shared libraries: {}'.format(wheel_libdir))
        else:
            _log('Not wheel distribution ({} not found)'.format(
                wheel_libdir))
            wheel_libdir = None

        if (3, 8) <= sys.version_info:
            if cuda_bin_path is not None:
                _log('Adding DLL search path: {}'.format(cuda_bin_path))
                os.add_dll_directory(cuda_bin_path)
            if wheel_libdir is not None:
                _log('Adding DLL search path: {}'.format(wheel_libdir))
                os.add_dll_directory(wheel_libdir)
        else:
            # Users are responsible for adding `%CUDA_PATH%/bin` to PATH.
            if wheel_libdir is not None:
                _log('Adding to PATH: {}'.format(wheel_libdir))
                path = os.environ.get('PATH', '')
                os.environ['PATH'] = wheel_libdir + os.pathsep + path


def get_cupy_install_path():
    # Path to the directory where the package is installed.
    return os.path.abspath(
        os.path.join(os.path.dirname(__file__), '..'))


def get_cupy_cuda_lib_path():
    """Returns the directory where CUDA external libraries are installed.

    This environment variable only affects wheel installations.

    Shared libraries are looked up from
    `$CUPY_CUDA_LIB_PATH/$CUDA_VER/$LIB_NAME/$LIB_VER/{lib,lib64,bin}`,
    e.g., `~/.cupy/cuda_lib/11.2/cudnn/8.1.1/lib64/libcudnn.so.8.1.1`.

    The default $CUPY_CUDA_LIB_PATH is `~/.cupy/cuda_lib`.
    """
    cupy_cuda_lib_path = os.environ.get('CUPY_CUDA_LIB_PATH', None)
    if cupy_cuda_lib_path is None:
        return os.path.expanduser('~/.cupy/cuda_lib')
    return os.path.abspath(cupy_cuda_lib_path)


def get_preload_config():
    global _preload_config
    if _preload_config is None:
        config_path = os.path.join(
            get_cupy_install_path(), 'cupy', '.data', '_wheel.json')
        if not os.path.exists(config_path):
            return None
        with open(config_path) as f:
            _preload_config = json.load(f)
    return _preload_config


def _can_attempt_preload(lib: str) -> bool:
    """Returns if the preload can be attempted."""

    config = get_preload_config()
    if (config is None) or (config['packaging'] == 'conda'):
        # We don't do preload if CuPy is installed from Conda-Forge, as we
        # cannot guarantee the version pinned in _wheel.json, which is
        # encoded in config[lib]['filenames'], is always available on
        # Conda-Forge. See here for the configuration files used in
        # Conda-Forge distributions.
        # https://github.com/conda-forge/cupy-feedstock/blob/master/recipe/preload_config/
        _log(f'Cannot preload {lib} as this is not a wheel installation')
        return False

    if lib not in _preload_libs:
        raise AssertionError(f'Unknown preload library: {lib}')

    if lib not in config:
        _log(f'Preload {lib} not configured in wheel')
        return False

    if _preload_libs[lib] is not None:
        _log(f'Preload already attempted: {lib}')
        return False

    return True


def _preload_library(lib):
    """Preload dependent shared libraries.

    The preload configuration file (cupy/.data/_wheel.json) will be added
    during the wheel build process.
    """

    _log(f'Preloading triggered for library: {lib}')

    if not _can_attempt_preload(lib):
        return
    _preload_libs[lib] = {}

    config = get_preload_config()
    cuda_version = config['cuda']
    _log('CuPy wheel package built for CUDA {}'.format(cuda_version))

    cupy_cuda_lib_path = get_cupy_cuda_lib_path()
    _log('CuPy CUDA library directory: {}'.format(cupy_cuda_lib_path))

    version = config[lib]['version']
    filenames = config[lib]['filenames']
    for filename in filenames:
        _log(f'Looking for {lib} version {version} ({filename})')

        # "lib": cuTENSOR (Linux/Windows) / NCCL (Linux)
        # "lib64": cuDNN (Linux)
        # "bin": cuDNN (Windows)
        libpath_cands = [
            os.path.join(
                cupy_cuda_lib_path, config['cuda'], lib, version, x,
                filename)
            for x in ['lib', 'lib64', 'bin']]
        for libpath in libpath_cands:
            if not os.path.exists(libpath):
                _log('Rejected candidate (not found): {}'.format(libpath))
                continue

            try:
                _log(f'Trying to load {libpath}')
                # Keep reference to the preloaded module.
                _preload_libs[lib][libpath] = ctypes.CDLL(libpath)
                _log('Loaded')
                break
            except Exception as e:
                e_type = type(e).__name__  # NOQA
                msg = (
                    f'CuPy failed to preload library ({libpath}): '
                    f'{e_type} ({e})')
                _log(msg)
                warnings.warn(msg)
        else:
            _log('File {} could not be found'.format(filename))

            # Lookup library with fully-qualified version (e.g.,
            # `libcudnn.so.X.Y.Z`).
            _log(f'Trying to load {filename} from default search path')
            try:
                _preload_libs[lib][filename] = ctypes.CDLL(filename)
                _log('Loaded')
            except Exception as e:
                # Fallback to the standard shared library lookup which only
                # uses the major version (e.g., `libcudnn.so.X`).
                _log(f'Library {lib} could not be preloaded: {e}')


def _get_preload_logs():
    return '\n'.join(_preload_logs)


def _preload_warning(lib, exc):
    config = get_preload_config()
    if config is not None and lib in config:
        msg = '''
{lib} library could not be loaded.

Reason: {exc_type} ({exc})

You can install the library by:
'''
        if config['packaging'] == 'pip':
            msg += '''
  $ python -m cupyx.tools.install_library --library {lib} --cuda {cuda}
'''
        elif config['packaging'] == 'conda':
            msg += '''
  $ conda install -c conda-forge {lib}
'''
        else:
            raise AssertionError
        msg = msg.format(
            lib=lib, exc_type=type(exc).__name__, exc=str(exc),
            cuda=config['cuda'])
        warnings.warn(msg)


def _detect_duplicate_installation():
    # importlib.metadata only available in Python 3.8+.
    if sys.version_info < (3, 8):
        return
    import importlib.metadata

    # List of all CuPy packages, including out-dated ones.
    known = [
        'cupy',
        'cupy-cuda80',
        'cupy-cuda90',
        'cupy-cuda91',
        'cupy-cuda92',
        'cupy-cuda100',
        'cupy-cuda101',
        'cupy-cuda102',
        'cupy-cuda110',
        'cupy-cuda111',
        'cupy-cuda112',
        'cupy-cuda113',
        'cupy-cuda114',
        'cupy-cuda115',
        'cupy-rocm-4-0',
        'cupy-rocm-4-1',
        'cupy-rocm-4-2',
        'cupy-rocm-4-3',
    ]
    cupy_installed = [
        name for name in known
        if list(importlib.metadata.distributions(name=name))]
    if 1 < len(cupy_installed):
        cupy_packages_list = ', '.join(sorted(cupy_installed))
        warnings.warn(f'''
--------------------------------------------------------------------------------

  CuPy may not function correctly because multiple CuPy packages are installed
  in your environment:

    {cupy_packages_list}

  Follow these steps to resolve this issue:

    1. For all packages listed above, run the following command to remove all
       existing CuPy installations:

         $ pip uninstall <package_name>

      If you previously installed CuPy via conda, also run the following:

         $ conda uninstall cupy

    2. Install the appropriate CuPy package.
       Refer to the Installation Guide for detailed instructions.

         https://docs.cupy.dev/en/stable/install.html

--------------------------------------------------------------------------------
''')


def _diagnose_import_error() -> str:
    # TODO(kmaehashi): provide better diagnostics.
    return '''\
Failed to import CuPy.

If you installed CuPy via wheels (cupy-cudaXXX or cupy-rocm-X-X), make sure that the package matches with the version of CUDA or ROCm installed.

On Linux, you may need to set LD_LIBRARY_PATH environment variable depending on how you installed CUDA/ROCm.
On Windows, try setting CUDA_PATH environment variable.

Check the Installation Guide for details:
  https://docs.cupy.dev/en/latest/install.html'''  # NOQA



import cupy
from cupy.cuda import nccl
from cupyx.distributed import _store
from cupyx.distributed._comm import _Backend


if nccl.available:
    # types are not compliant with windows on long/int32 issue
    # but nccl does not support windows so we don't care
    _nccl_dtypes = {'b': nccl.NCCL_INT8,
                    'B': nccl.NCCL_UINT8,
                    'i': nccl.NCCL_INT32,
                    'I': nccl.NCCL_UINT32,
                    'l': nccl.NCCL_INT64,
                    'L': nccl.NCCL_UINT64,
                    'q': nccl.NCCL_INT64,
                    'Q': nccl.NCCL_UINT64,
                    'e': nccl.NCCL_FLOAT16,
                    'f': nccl.NCCL_FLOAT32,
                    'd': nccl.NCCL_FLOAT64,
                    # Size of array will be doubled
                    'F': nccl.NCCL_FLOAT32,
                    'D': nccl.NCCL_FLOAT64}

    _nccl_ops = {'sum': nccl.NCCL_SUM,
                 'prod': nccl.NCCL_PROD,
                 'max': nccl.NCCL_MAX,
                 'min': nccl.NCCL_MIN}
else:
    _nccl_dtypes = {}

    _nccl_ops = {}


class NCCLBackend(_Backend):
    """Interface that uses NVIDIA's NCCL to perform communications.

    Args:
        n_devices (int): Total number of devices that will be used in the
            distributed execution.
        rank (int): Unique id of the GPU that the communicator is associated to
            its value needs to be `0 <= rank < n_devices`.
        host (str, optional): host address for the process rendezvous on
            initialization. Defaults to `"127.0.0.1"`.
        port (int, optional): port used for the process rendezvous on
            initialization. Defaults to `13333`.
    """

    def __init__(self, n_devices, rank,
                 host=_store._DEFAULT_HOST, port=_store._DEFAULT_PORT):
        super().__init__(n_devices, rank, host, port)
        if rank == 0:
            self._store.run(host, port)
            nccl_id = nccl.get_unique_id()
            # get_unique_id return negative values due to cython issues
            # with bytes && c strings. We shift them by 128 to
            # avoid issues
            nccl_id = bytes([b + 128 for b in nccl_id])
            self._store_proxy['nccl_id'] = nccl_id
            self._store_proxy.barrier()
        else:
            self._store_proxy.barrier()
            nccl_id = self._store_proxy['nccl_id']
        # Initialize devices
        nccl_id = tuple([int(b) - 128 for b in nccl_id])
        self._comm = nccl.NcclCommunicator(n_devices, nccl_id, rank)

    def _check_contiguous(self, array):
        if not array.flags.c_contiguous and not array.flags.f_contiguous:
            raise RuntimeError(
                'NCCL requires arrays to be either c- or f-contiguous')

    def _get_nccl_dtype_and_count(self, array, count=None):
        dtype = array.dtype.char
        if dtype not in _nccl_dtypes:
            raise TypeError(f'Unknown dtype {array.dtype} for NCCL')
        nccl_dtype = _nccl_dtypes[dtype]
        if count is None:
            count = array.size
        if dtype in 'FD':
            return nccl_dtype, 2 * count
        return nccl_dtype, count

    def _get_stream(self, stream):
        if stream is None:
            stream = cupy.cuda.stream.get_current_stream()
        return stream.ptr

    def _get_op(self, op, dtype):
        if op not in _nccl_ops:
            raise RuntimeError(f'Unknown op {op} for NCCL')
        if dtype in 'FD' and op != nccl.NCCL_SUM:
            raise ValueError(
                'Only nccl.SUM is supported for complex arrays')
        return _nccl_ops[op]

    def all_reduce(self, in_array, out_array, op='sum', stream=None):
        """Performs an all reduce operation.

        Args:
            in_array (cupy.ndarray): array to be sent.
            out_array (cupy.ndarray): array where the result with be stored.
            op (str): reduction operation, can be one of
                ('sum', 'prod', 'min' 'max'), arrays of complex type only
                support `'sum'`. Defaults to `'sum'`.
            stream (cupy.cuda.Stream, optional): if supported, stream to
                perform the communication.
        """
        self._check_contiguous(in_array)
        self._check_contiguous(out_array)
        stream = self._get_stream(stream)
        dtype, count = self._get_nccl_dtype_and_count(in_array)
        op = self._get_op(op, in_array.dtype.char)
        self._comm.allReduce(
            in_array.data.ptr, out_array.data.ptr, count, dtype, op, stream)

    def reduce(self, in_array, out_array, root=0, op='sum', stream=None):
        """Performs a reduce operation.

        Args:
            in_array (cupy.ndarray): array to be sent.
            out_array (cupy.ndarray): array where the result with be stored.
                will only be modified by the `root` process.
            root (int, optional): rank of the process that will perform the
                reduction. Defaults to `0`.
            op (str): reduction operation, can be one of
                ('sum', 'prod', 'min' 'max'), arrays of complex type only
                support `'sum'`. Defaults to `'sum'`.
            stream (cupy.cuda.Stream, optional): if supported, stream to
                perform the communication.
        """
        self._check_contiguous(in_array)
        if self.rank == root:
            self._check_contiguous(out_array)
        stream = self._get_stream(stream)
        dtype, count = self._get_nccl_dtype_and_count(in_array)
        op = self._get_op(op, in_array.dtype.char)
        self._comm.reduce(
            in_array.data.ptr, out_array.data.ptr,
            count, dtype, op, root, stream)

    def broadcast(self, in_out_array, root=0, stream=None):
        """Performs a broadcast operation.

        Args:
            in_out_array (cupy.ndarray): array to be sent for `root` rank.
                Other ranks will receive the broadcast data here.
            root (int, optional): rank of the process that will send the
                broadcast. Defaults to `0`.
            stream (cupy.cuda.Stream, optional): if supported, stream to
                perform the communication.
        """
        # in_out_array for rank !=0 will be used as output
        self._check_contiguous(in_out_array)
        stream = self._get_stream(stream)
        dtype, count = self._get_nccl_dtype_and_count(in_out_array)
        self._comm.broadcast(
            in_out_array.data.ptr, in_out_array.data.ptr,
            count, dtype, root, stream)

    def reduce_scatter(
            self, in_array, out_array, count, op='sum', stream=None):
        """Performs a reduce scatter operation.

        Args:
            in_array (cupy.ndarray): array to be sent.
            out_array (cupy.ndarray): array where the result with be stored.
            count (int): Number of elements to send to each rank.
            op (str): reduction operation, can be one of
                ('sum', 'prod', 'min' 'max'), arrays of complex type only
                support `'sum'`. Defaults to `'sum'`.
            stream (cupy.cuda.Stream, optional): if supported, stream to
                perform the communication.
        """
        self._check_contiguous(in_array)
        self._check_contiguous(out_array)
        stream = self._get_stream(stream)
        dtype, count = self._get_nccl_dtype_and_count(in_array, count)
        op = self._get_op(op, in_array.dtype.char)
        self._comm.reduceScatter(
            in_array.data.ptr, out_array.data.ptr, count, dtype, op, stream)

    def all_gather(self, in_array, out_array, count, stream=None):
        """Performs an all gather operation.

        Args:
            in_array (cupy.ndarray): array to be sent.
            out_array (cupy.ndarray): array where the result with be stored.
            count (int): Number of elements to send to each rank.
            op (str): reduction operation, can be one of
                ('sum', 'prod', 'min' 'max'), arrays of complex type only
                support `'sum'`. Defaults to `'sum'`.
            stream (cupy.cuda.Stream, optional): if supported, stream to
                perform the communication.
        """
        self._check_contiguous(in_array)
        self._check_contiguous(out_array)
        stream = self._get_stream(stream)
        dtype, count = self._get_nccl_dtype_and_count(in_array, count)
        self._comm.allGather(
            in_array.data.ptr, out_array.data.ptr, count, dtype, stream)

    def send(self, array, peer, stream=None):
        """Performs a send operation.

        Args:
            array (cupy.ndarray): array to be sent.
            peer (int): rank of the process `array` will be sent to.
            stream (cupy.cuda.Stream, optional): if supported, stream to
                perform the communication.
        """
        self._check_contiguous(array)
        stream = self._get_stream(stream)
        dtype, count = self._get_nccl_dtype_and_count(array)
        self._send(array, peer, dtype, count, stream)

    def _send(self, array, peer, dtype, count, stream=None):
        self._comm.send(array.data.ptr, count, dtype, peer, stream)

    def recv(self, out_array, peer, stream=None):
        """Performs a receive operation.

        Args:
            array (cupy.ndarray): array used to receive data.
            peer (int): rank of the process `array` will be received from.
            stream (cupy.cuda.Stream, optional): if supported, stream to
                perform the communication.
        """
        self._check_contiguous(out_array)
        stream = self._get_stream(stream)
        dtype, count = self._get_nccl_dtype_and_count(out_array)
        self._recv(out_array, peer, dtype, count, stream)

    def _recv(self, out_array, peer, dtype, count, stream=None):
        self._comm.recv(out_array.data.ptr, count, dtype, peer, stream)

    # TODO(ecastill) implement nccl missing calls combining the above ones
    # AlltoAll, AllGather, and similar MPI calls that can be easily implemented
    def send_recv(self, in_array, out_array, peer, stream=None):
        """Performs a send and receive operation.

        Args:
            in_array (cupy.ndarray): array to be sent.
            out_array (cupy.ndarray): array used to receive data.
            peer (int): rank of the process to send `in_array` and receive
                `out_array`.
            stream (cupy.cuda.Stream, optional): if supported, stream to
                perform the communication.
        """
        self._check_contiguous(in_array)
        self._check_contiguous(out_array)
        stream = self._get_stream(stream)
        idtype, icount = self._get_nccl_dtype_and_count(in_array)
        odtype, ocount = self._get_nccl_dtype_and_count(out_array)
        nccl.groupStart()
        self._send(in_array, peer, idtype, icount, stream)
        self._recv(out_array, peer, odtype, ocount, stream)
        nccl.groupEnd()

    def scatter(self, in_array, out_array, root=0, stream=None):
        """Performs a scatter operation.

        Args:
            in_array (cupy.ndarray): array to be sent. Its shape must be
                `(total_ranks, ...)`.
            out_array (cupy.ndarray): array where the result with be stored.
            root (int): rank that will send the `in_array` to other ranks.
            stream (cupy.cuda.Stream, optional): if supported, stream to
                perform the communication.
        """
        if in_array.shape[0] != self._n_devices:
            raise RuntimeError(
                f'scatter requires in_array to have {self._n_devices}'
                f'elements in its first dimension, found {in_array.shape}')
        self._check_contiguous(in_array)
        self._check_contiguous(out_array)
        stream = self._get_stream(stream)
        nccl.groupStart()
        if root == self.rank:
            for i in range(self._n_devices):
                array = in_array[i]
                idtype, icount = self._get_nccl_dtype_and_count(array)
                self._send(array, i, idtype, icount, stream)
        dtype, count = self._get_nccl_dtype_and_count(out_array)
        self._recv(out_array, root, dtype, count, stream)
        nccl.groupEnd()

    def gather(self, in_array, out_array, root=0, stream=None):
        """Performs a gather operation.

        Args:
            in_array (cupy.ndarray): array to be sent.
            out_array (cupy.ndarray): array where the result with be stored.
                Its shape must be `(total_ranks, ...)`.
            root (int): rank that will receive `in_array` from other ranks.
            stream (cupy.cuda.Stream, optional): if supported, stream to
                perform the communication.
        """
        # TODO(ecastill) out_array needs to have comm size in shape[0]
        if out_array.shape[0] != self._n_devices:
            raise RuntimeError(
                f'gather requires out_array to have {self._n_devices}'
                f'elements in its first dimension, found {out_array.shape}')
        self._check_contiguous(in_array)
        self._check_contiguous(out_array)
        stream = self._get_stream(stream)
        nccl.groupStart()
        if root == self.rank:
            for i in range(self._n_devices):
                array = out_array[i]
                odtype, ocount = self._get_nccl_dtype_and_count(array)
                self._recv(array, i, odtype, ocount, stream)
        dtype, count = self._get_nccl_dtype_and_count(in_array)
        self._send(in_array, root, dtype, count, stream)
        nccl.groupEnd()

    def all_to_all(self, in_array, out_array, stream=None):
        """Performs an all to all operation.

        Args:
            in_array (cupy.ndarray): array to be sent. Its shape must be
                `(total_ranks, ...)`.
            out_array (cupy.ndarray): array where the result with be stored.
                Its shape must be `(total_ranks, ...)`.
            stream (cupy.cuda.Stream, optional): if supported, stream to
                perform the communication.
        """
        # TODO(ecastill) out_array needs to have comm size in shape[0]
        if out_array.shape[0] != self._n_devices:
            raise RuntimeError(
                f'all_to_all requires in_array to have {self._n_devices}'
                f'elements in its first dimension, found {in_array.shape}')
        if out_array.shape[0] != self._n_devices:
            raise RuntimeError(
                f'all_to_all requires out_array to have {self._n_devices}'
                f'elements in its first dimension, found {out_array.shape}')
        self._check_contiguous(in_array)
        self._check_contiguous(out_array)
        stream = self._get_stream(stream)
        idtype, icount = self._get_nccl_dtype_and_count(in_array[0])
        odtype, ocount = self._get_nccl_dtype_and_count(out_array[0])
        # TODO check out dtypes are the same as in dtypes
        nccl.groupStart()
        for i in range(self._n_devices):
            self._send(in_array[i], i, idtype, icount, stream)
            self._recv(out_array[i], i, odtype, ocount, stream)
        nccl.groupEnd()

    def barrier(self):
        """Performs a barrier operation.

        The barrier is done in the cpu and is a explicit synchronization
        mechanism that halts the thread progression.
        """
        # implements a barrier CPU side
        # TODO allow multiple barriers to be executed
        self._store_proxy.barrier()



import unittest

import numpy
try:
    import scipy.sparse
    scipy_available = True
except ImportError:
    scipy_available = False

import cupy
from cupy import testing
from cupyx.scipy import sparse


@testing.parameterize(*testing.product({
    'shape': [(8, 3), (4, 4), (3, 8)],
    'a_format': ['dense', 'csr', 'csc', 'coo'],
    'out_format': [None, 'csr', 'csc'],
}))
@unittest.skipUnless(scipy_available, 'requires scipy')
class TestExtract(unittest.TestCase):

    density = 0.75

    def _make_matrix(self, dtype):
        a = testing.shaped_random(self.shape, numpy, dtype=dtype)
        a[a > self.density] = 0
        b = cupy.array(a)
        if self.a_format == 'csr':
            a = scipy.sparse.csr_matrix(a)
            b = sparse.csr_matrix(b)
        elif self.a_format == 'csc':
            a = scipy.sparse.csc_matrix(a)
            b = sparse.csc_matrix(b)
        elif self.a_format == 'coo':
            a = scipy.sparse.coo_matrix(a)
            b = sparse.coo_matrix(b)
        return a, b

    @testing.for_dtypes('fdFD')
    def test_tril(self, dtype):
        np_a, cp_a = self._make_matrix(dtype)
        m, n = self.shape
        for k in range(-m+1, n):
            np_out = scipy.sparse.tril(np_a, k=k, format=self.out_format)
            cp_out = sparse.tril(cp_a, k=k, format=self.out_format)
            assert np_out.format == cp_out.format
            assert np_out.nnz == cp_out.nnz
            cupy.testing.assert_allclose(np_out.todense(), cp_out.todense())

    @testing.for_dtypes('fdFD')
    def test_triu(self, dtype):
        np_a, cp_a = self._make_matrix(dtype)
        m, n = self.shape
        for k in range(-m+1, n):
            np_out = scipy.sparse.triu(np_a, k=k, format=self.out_format)
            cp_out = sparse.triu(cp_a, k=k, format=self.out_format)
            assert np_out.format == cp_out.format
            assert np_out.nnz == cp_out.nnz
            cupy.testing.assert_allclose(np_out.todense(), cp_out.todense())

    @testing.for_dtypes('fdFD')
    def test_find(self, dtype):
        if self.out_format is not None:
            unittest.SkipTest()
        np_a, cp_a = self._make_matrix(dtype)
        np_row, np_col, np_data = scipy.sparse.find(np_a)
        cp_row, cp_col, cp_data = sparse.find(cp_a)
        # Note: Check the results by reconstructing the sparse matrix from the
        # results of find, as SciPy and CuPy differ in the data order.
        np_out = scipy.sparse.coo_matrix((np_data, (np_row, np_col)),
                                         shape=self.shape)
        cp_out = sparse.coo_matrix((cp_data, (cp_row, cp_col)),
                                   shape=self.shape)
        cupy.testing.assert_allclose(np_out.todense(), cp_out.todense())



import unittest

import numpy
try:
    import scipy.sparse  # NOQA
    import scipy.sparse.csgraph  # NOQA
    scipy_available = True
except ImportError:
    scipy_available = False
import cupyx.scipy.sparse.csgraph  # NOQA
try:
    import pylibcugraph  # NOQA
    pylibcugraph_available = True
except ImportError:
    pylibcugraph_available = False
from cupy import testing


@testing.parameterize(*testing.product({
    'dtype': ['float32', 'float64'],
    'm': [5, 10, 20],
    'nnz_per_row': [1, 2, 3],
    'directed': [True, False],
    'connection': ['weak', 'strong'],
    'return_labels': [True, False],
}))
@unittest.skipUnless(scipy_available and pylibcugraph_available,
                     'requires scipy and pylibcugraph')
class TestConnectedComponents(unittest.TestCase):

    def _make_matrix(self, dtype, xp):
        shape = (self.m, self.m)
        density = self.nnz_per_row / self.m
        a = testing.shaped_random(shape, xp, dtype=dtype, scale=1)
        a = a / density
        a[a > 1] = 0
        return a

    @testing.numpy_cupy_array_equal(sp_name='sp')
    def test_connected_components(self, xp, sp):
        a = self._make_matrix(self.dtype, xp)
        a = sp.csr_matrix(a)
        if self.return_labels:
            n, labels = sp.csgraph.connected_components(
                a, directed=self.directed, connection=self.connection,
                return_labels=self.return_labels)
            if xp == numpy and self.directed and self.connection == 'strong':
                # Note: Label IDs are adjusted as SciPy returns un-ordered
                # labels in strong connection case.
                table = xp.zeros((n,), dtype=numpy.int32) - 1
                j = 0
                for i in range(labels.size):
                    if table[labels[i]] < 0:
                        table[labels[i]] = j
                        j = j + 1
                    labels[i] = table[labels[i]]
            return n, labels
        else:
            return sp.csgraph.connected_components(
                a, directed=self.directed, connection=self.connection,
                return_labels=self.return_labels)



import sys
import time

import cupy
from cupy import cuda
from cupy.cuda import nccl
from cupy import testing

from cupyx.distributed import init_process_group
from cupyx.distributed._nccl_comm import NCCLBackend
from cupyx.distributed._store import ExceptionAwareProcess

nccl_available = nccl.available


N_WORKERS = 2


def _launch_workers(n_workers, func, args=()):
    processes = []
    # TODO catch exceptions
    for rank in range(n_workers):
        p = ExceptionAwareProcess(
            target=func,
            args=(rank, n_workers) + args)
        p.start()
        processes.append(p)

    for p in processes:
        p.join()


def broadcast(dtype):
    if dtype in 'hH':
        return  # nccl does not support int16

    def run_broadcast(rank, n_workers, root, dtype):
        dev = cuda.Device(rank)
        dev.use()
        comm = NCCLBackend(n_workers, rank)
        expected = cupy.arange(2 * 3 * 4, dtype=dtype).reshape((2, 3, 4))
        if rank == root:
            in_array = expected
        else:
            in_array = cupy.zeros((2, 3, 4), dtype=dtype)
        comm.broadcast(in_array, root)
        testing.assert_allclose(in_array, expected)

    _launch_workers(N_WORKERS, run_broadcast, (0, dtype))
    _launch_workers(N_WORKERS, run_broadcast, (1, dtype))


def reduce(dtype):
    if dtype in 'hH':
        return  # nccl does not support int16

    def run_reduce(rank, n_workers, root, dtype):
        dev = cuda.Device(rank)
        dev.use()
        comm = NCCLBackend(n_workers, rank)
        in_array = cupy.arange(2 * 3 * 4, dtype='f').reshape(2, 3, 4)
        out_array = cupy.zeros((2, 3, 4), dtype='f')
        comm.reduce(in_array, out_array, root)
        if rank == root:
            testing.assert_allclose(out_array, 2 * in_array)

    _launch_workers(N_WORKERS, run_reduce, (0, dtype))
    _launch_workers(N_WORKERS, run_reduce, (1, dtype))


def all_reduce(dtype):
    if dtype in 'hH':
        return  # nccl does not support int16

    def run_all_reduce(rank, n_workers, dtype):
        dev = cuda.Device(rank)
        dev.use()
        comm = NCCLBackend(n_workers, rank)
        in_array = cupy.arange(2 * 3 * 4, dtype='f').reshape(2, 3, 4)
        out_array = cupy.zeros((2, 3, 4), dtype='f')

        comm.all_reduce(in_array, out_array)
        testing.assert_allclose(out_array, 2 * in_array)

    _launch_workers(N_WORKERS, run_all_reduce, (dtype,))


def reduce_scatter(dtype):
    if dtype in 'hH':
        return  # nccl does not support int16

    def run_reduce_scatter(rank, n_workers, dtype):
        dev = cuda.Device(rank)
        dev.use()
        comm = NCCLBackend(n_workers, rank)
        in_array = 1 + cupy.arange(
            n_workers * 10, dtype='f').reshape(n_workers, 10)
        out_array = cupy.zeros((10,), dtype='f')

        comm.reduce_scatter(in_array, out_array, 10)
        testing.assert_allclose(out_array, 2 * in_array[rank])

    _launch_workers(N_WORKERS, run_reduce_scatter, (dtype,))


def all_gather(dtype):
    if dtype in 'hH':
        return  # nccl does not support int16

    def run_all_gather(rank, n_workers, dtype):
        dev = cuda.Device(rank)
        dev.use()
        comm = NCCLBackend(n_workers, rank)
        in_array = (rank + 1) * cupy.arange(
            n_workers * 10, dtype='f').reshape(n_workers, 10)
        out_array = cupy.zeros((n_workers, 10), dtype='f')
        comm.all_gather(in_array, out_array, 10)
        expected = 1 + cupy.arange(n_workers).reshape(n_workers, 1)
        expected = expected * cupy.broadcast_to(
            cupy.arange(10, dtype='f'), (n_workers, 10))
        testing.assert_allclose(out_array, expected)

    _launch_workers(N_WORKERS, run_all_gather, (dtype,))


def send_and_recv(dtype):
    if dtype in 'hH':
        return  # nccl does not support int16

    def run_send_and_recv(rank, n_workers, dtype):
        dev = cuda.Device(rank)
        dev.use()
        comm = NCCLBackend(n_workers, rank)
        in_array = cupy.arange(10, dtype='f')
        out_array = cupy.zeros((10,), dtype='f')
        if rank == 0:
            comm.send(in_array, 1)
        else:
            comm.recv(out_array, 0)
            testing.assert_allclose(out_array, in_array)

    _launch_workers(N_WORKERS, run_send_and_recv, (dtype,))


def send_recv(dtype):
    if dtype in 'hH':
        return  # nccl does not support int16

    def run_send_recv(rank, n_workers, dtype):
        dev = cuda.Device(rank)
        dev.use()
        comm = NCCLBackend(n_workers, rank)
        in_array = cupy.arange(10, dtype='f')
        for i in range(n_workers):
            out_array = cupy.zeros((10,), dtype='f')
            comm.send_recv(in_array, out_array, i)
            testing.assert_allclose(out_array, in_array)

    _launch_workers(N_WORKERS, run_send_recv, (dtype,))


def scatter(dtype):
    if dtype in 'hH':
        return  # nccl does not support int16

    def run_scatter(rank, n_workers, root, dtype):
        dev = cuda.Device(rank)
        dev.use()
        comm = NCCLBackend(n_workers, rank)
        in_array = 1 + cupy.arange(
            n_workers * 10, dtype='f').reshape(n_workers, 10)
        out_array = cupy.zeros((10,), dtype='f')

        comm.scatter(in_array, out_array, root)
        if rank > 0:
            testing.assert_allclose(out_array, in_array[rank])

    _launch_workers(N_WORKERS, run_scatter, (0, dtype))
    _launch_workers(N_WORKERS, run_scatter, (1, dtype))


def gather(dtype):
    if dtype in 'hH':
        return  # nccl does not support int16

    def run_gather(rank, n_workers, root, dtype):
        dev = cuda.Device(rank)
        dev.use()
        comm = NCCLBackend(n_workers, rank)
        in_array = (rank + 1) * cupy.arange(10, dtype='f')
        out_array = cupy.zeros((n_workers, 10), dtype='f')
        comm.gather(in_array, out_array, root)
        if rank == root:
            expected = 1 + cupy.arange(n_workers).reshape(n_workers, 1)
            expected = expected * cupy.broadcast_to(
                cupy.arange(10, dtype='f'), (n_workers, 10))
            testing.assert_allclose(out_array, expected)

    _launch_workers(N_WORKERS, run_gather, (0, dtype))
    _launch_workers(N_WORKERS, run_gather, (1, dtype))


def all_to_all(dtype):
    if dtype in 'hH':
        return  # nccl does not support int16

    def run_all_to_all(rank, n_workers, dtype):
        dev = cuda.Device(rank)
        dev.use()
        comm = NCCLBackend(n_workers, rank)
        in_array = cupy.arange(
            n_workers * 10, dtype='f').reshape(n_workers, 10)
        out_array = cupy.zeros((n_workers, 10), dtype='f')
        comm.all_to_all(in_array, out_array)
        expected = (10 * rank) + cupy.broadcast_to(
            cupy.arange(10, dtype='f'), (n_workers, 10))
        testing.assert_allclose(out_array, expected)

    _launch_workers(N_WORKERS, run_all_to_all, (dtype,))


def barrier():
    def run_barrier(rank, n_workers):
        dev = cuda.Device(rank)
        dev.use()
        comm = NCCLBackend(n_workers, rank)
        comm.barrier()
        before = time.time()
        if rank == 0:
            time.sleep(2)
        comm.barrier()
        after = time.time()
        assert int(after - before) == 2

    _launch_workers(N_WORKERS, run_barrier)


def init():
    def run_init(rank, n_workers):
        dev = cuda.Device(rank)
        dev.use()
        comm = init_process_group(n_workers, rank)
        # Do a simple call to verify we got a valid comm
        in_array = cupy.zeros(1)
        if rank == 0:
            in_array = in_array + 1
        comm.broadcast(in_array, 0)
        testing.assert_allclose(in_array, cupy.ones(1))

    _launch_workers(N_WORKERS, run_init)


if __name__ == '__main__':
    # Run the templatized test
    func = globals()[sys.argv[1]]
    # dtype is the char representation
    dtype = sys.argv[2] if len(sys.argv) == 3 else None
    if dtype is not None:
        func(dtype)
    else:
        func()



import cupy
from cupy import _core
from cupy._core import fusion
from cupy import _util

from cupy._core import _routines_indexing as _indexing
from cupy._core import _routines_statistics as _statistics


def argmax(a, axis=None, dtype=None, out=None, keepdims=False):
    """Returns the indices of the maximum along an axis.

    Args:
        a (cupy.ndarray): Array to take argmax.
        axis (int): Along which axis to find the maximum. ``a`` is flattened by
            default.
        dtype: Data type specifier.
        out (cupy.ndarray): Output array.
        keepdims (bool): If ``True``, the axis ``axis`` is preserved as an axis
            of length one.

    Returns:
        cupy.ndarray: The indices of the maximum of ``a`` along an axis.

    .. note::
       ``dtype`` and ``keepdim`` arguments are specific to CuPy. They are
       not in NumPy.

    .. note::
       ``axis`` argument accepts a tuple of ints, but this is specific to
       CuPy. NumPy does not support it.

    .. seealso:: :func:`numpy.argmax`

    """
    # TODO(okuta): check type
    return a.argmax(axis=axis, dtype=dtype, out=out, keepdims=keepdims)


def nanargmax(a, axis=None, dtype=None, out=None, keepdims=False):
    """Return the indices of the maximum values in the specified axis ignoring
    NaNs. For all-NaN slice ``-1`` is returned.
    Subclass cannot be passed yet, subok=True still unsupported

    Args:
        a (cupy.ndarray): Array to take nanargmax.
        axis (int): Along which axis to find the maximum. ``a`` is flattened by
            default.

    Returns:
        cupy.ndarray: The indices of the maximum of ``a``
        along an axis ignoring NaN values.

    .. note:: For performance reasons, ``cupy.nanargmax`` returns
            ``out of range values`` for all-NaN slice
            whereas ``numpy.nanargmax`` raises ``ValueError``
    .. seealso:: :func:`numpy.nanargmax`
    """
    if a.dtype.kind in 'biu':
        return argmax(a, axis=axis)

    return _statistics._nanargmax(a, axis, dtype, out, keepdims)


def argmin(a, axis=None, dtype=None, out=None, keepdims=False):
    """Returns the indices of the minimum along an axis.

    Args:
        a (cupy.ndarray): Array to take argmin.
        axis (int): Along which axis to find the minimum. ``a`` is flattened by
            default.
        dtype: Data type specifier.
        out (cupy.ndarray): Output array.
        keepdims (bool): If ``True``, the axis ``axis`` is preserved as an axis
            of length one.

    Returns:
        cupy.ndarray: The indices of the minimum of ``a`` along an axis.

    .. note::
       ``dtype`` and ``keepdim`` arguments are specific to CuPy. They are
       not in NumPy.

    .. note::
       ``axis`` argument accepts a tuple of ints, but this is specific to
       CuPy. NumPy does not support it.

    .. seealso:: :func:`numpy.argmin`

    """
    # TODO(okuta): check type
    return a.argmin(axis=axis, dtype=dtype, out=out, keepdims=keepdims)


def nanargmin(a, axis=None, dtype=None, out=None, keepdims=False):
    """Return the indices of the minimum values in the specified axis ignoring
    NaNs. For all-NaN slice ``-1`` is returned.
    Subclass cannot be passed yet, subok=True still unsupported

    Args:
        a (cupy.ndarray): Array to take nanargmin.
        axis (int): Along which axis to find the minimum. ``a`` is flattened by
            default.

    Returns:
        cupy.ndarray: The indices of the minimum of ``a``
        along an axis ignoring NaN values.

    .. note:: For performance reasons, ``cupy.nanargmin`` returns
            ``out of range values`` for all-NaN slice
            whereas ``numpy.nanargmin`` raises ``ValueError``
    .. seealso:: :func:`numpy.nanargmin`
    """
    if a.dtype.kind in 'biu':
        return argmin(a, axis=axis)

    return _statistics._nanargmin(a, axis, dtype, out, keepdims)


def nonzero(a):
    """Return the indices of the elements that are non-zero.

    Returns a tuple of arrays, one for each dimension of a,
    containing the indices of the non-zero elements in that dimension.

    Args:
        a (cupy.ndarray): array

    Returns:
        tuple of arrays: Indices of elements that are non-zero.

    .. warning::

        This function may synchronize the device.

    .. seealso:: :func:`numpy.nonzero`

    """
    _util.check_array(a, arg_name='a')
    return a.nonzero()


def flatnonzero(a):
    """Return indices that are non-zero in the flattened version of a.

    This is equivalent to a.ravel().nonzero()[0].

    Args:
        a (cupy.ndarray): input array

    Returns:
        cupy.ndarray: Output array,
        containing the indices of the elements of a.ravel() that are non-zero.

    .. warning::

        This function may synchronize the device.

    .. seealso:: :func:`numpy.flatnonzero`
    """
    _util.check_array(a, arg_name='a')
    return a.ravel().nonzero()[0]


_where_ufunc = _core.create_ufunc(
    'cupy_where',
    ('???->?', '?bb->b', '?BB->B', '?hh->h', '?HH->H', '?ii->i', '?II->I',
     '?ll->l', '?LL->L', '?qq->q', '?QQ->Q', '?ee->e', '?ff->f',
     # On CUDA 6.5 these combinations don't work correctly (on CUDA >=7.0, it
     # works).
     # See issue #551.
     '?hd->d', '?Hd->d',
     '?dd->d', '?FF->F', '?DD->D'),
    'out0 = in0 ? in1 : in2')


def where(condition, x=None, y=None):
    """Return elements, either from x or y, depending on condition.

    If only condition is given, return ``condition.nonzero()``.

    Args:
        condition (cupy.ndarray): When True, take x, otherwise take y.
        x (cupy.ndarray): Values from which to choose on ``True``.
        y (cupy.ndarray): Values from which to choose on ``False``.

    Returns:
        cupy.ndarray: Each element of output contains elements of ``x`` when
        ``condition`` is ``True``, otherwise elements of ``y``. If only
        ``condition`` is given, return the tuple ``condition.nonzero()``,
        the indices where ``condition`` is True.

    .. warning::

        This function may synchronize the device if both ``x`` and ``y`` are
        omitted.

    .. seealso:: :func:`numpy.where`

    """

    missing = (x is None, y is None).count(True)

    if missing == 1:
        raise ValueError('Must provide both \'x\' and \'y\' or neither.')
    if missing == 2:
        return nonzero(condition)  # may synchronize

    if fusion._is_fusing():
        return fusion._call_ufunc(_where_ufunc, condition, x, y)
    return _where_ufunc(condition.astype('?'), x, y)


def argwhere(a):
    """Return the indices of the elements that are non-zero.

    Returns a (N, ndim) dimantional array containing the
    indices of the non-zero elements. Where `N` is number of
    non-zero elements and `ndim` is dimension of the given array.

    Args:
        a (cupy.ndarray): array

    Returns:
        cupy.ndarray: Indices of elements that are non-zero.

    .. seealso:: :func:`numpy.argwhere`

    """
    _util.check_array(a, arg_name='a')
    return _indexing._ndarray_argwhere(a)


# This is to allow using the same kernels for all dtypes, ints & floats
# as nan is a special case
_preamble = '''
template<typename T>
__device__ bool _isnan(T val) {
    return val != val;
}
'''


_hip_preamble = r'''
#ifdef __HIP_DEVICE_COMPILE__
  #define no_thread_divergence(do_work, to_return) \
    if (!is_done) {                                \
      do_work;                                     \
      is_done = true;                              \
    }
#else
  #define no_thread_divergence(do_work, to_return) \
    do_work;                                       \
    if (to_return) { return; }
#endif
'''


_searchsorted_kernel = _core.ElementwiseKernel(
    'S x, raw T bins, int64 n_bins, bool side_is_right, '
    'bool assume_increasing',
    'int64 y',
    '''
    #ifdef __HIP_DEVICE_COMPILE__
    bool is_done = false;
    #endif

    // Array is assumed to be monotonically
    // increasing unless a check is requested with the
    // `assume_increasing = False` parameter.
    // `digitize` allows increasing and decreasing arrays.
    bool inc = true;
    if (!assume_increasing && n_bins >= 2) {
        // In the case all the bins are nan the array is considered
        // to be decreasing in numpy
        inc = (bins[0] <= bins[n_bins-1])
              || (!_isnan<T>(bins[0]) && _isnan<T>(bins[n_bins-1]));
    }

    if (_isnan<S>(x)) {
        long long pos = (inc ? n_bins : 0);
        if (!side_is_right) {
            if (inc) {
                while (pos > 0 && _isnan<T>(bins[pos-1])) {
                    --pos;
                }
            } else {
                while (pos < n_bins && _isnan<T>(bins[pos])) {
                    ++pos;
                }
            }
        }
        no_thread_divergence( y = pos , true )
    }

    bool greater = false;
    if (side_is_right) {
        greater = inc && x >= bins[n_bins-1];
    } else {
        greater = (inc ? x > bins[n_bins-1] : x <= bins[n_bins-1]);
    }
    if (greater) {
        no_thread_divergence( y = n_bins , true )
    }

    long long left = 0;
    // In the case the bins is all NaNs, digitize
    // needs to place all the valid values to the right
    if (!inc) {
        while (_isnan<T>(bins[left]) && left < n_bins) {
            ++left;
        }
        if (left == n_bins) {
            no_thread_divergence( y = n_bins , true )
        }
        if (side_is_right
                && !_isnan<T>(bins[n_bins-1]) && !_isnan<S>(x)
                && bins[n_bins-1] > x) {
            no_thread_divergence( y = n_bins , true )
        }
    }

    long long right = n_bins-1;
    while (left < right) {
        long long m = left + (right - left) / 2;
        bool look_right = true;
        if (side_is_right) {
            look_right = (inc ? bins[m] <= x : bins[m] > x);
        } else {
            look_right = (inc ? bins[m] < x : bins[m] >= x);
        }
        if (look_right) {
            left = m + 1;
        } else {
            right = m;
        }
    }
    no_thread_divergence( y = right , false )
    ''', name='cupy_searchsorted_kernel', preamble=_preamble+_hip_preamble)


def searchsorted(a, v, side='left', sorter=None):
    """Finds indices where elements should be inserted to maintain order.

    Find the indices into a sorted array ``a`` such that,
    if the corresponding elements in ``v`` were inserted before the indices,
    the order of ``a`` would be preserved.

    Args:
        a (cupy.ndarray): Input array. If ``sorter`` is ``None``, then
            it must be sorted in ascending order,
            otherwise ``sorter`` must be an array of indices that sort it.
        v (cupy.ndarray): Values to insert into ``a``.
        side : {'left', 'right'}
            If ``left``, return the index of the first suitable location found
            If ``right``, return the last such index.
            If there is no suitable index, return either 0 or length of ``a``.
        sorter : 1-D array_like
            Optional array of integer indices that sort array ``a`` into
            ascending order. They are typically the result of
            :func:`~cupy.argsort`.

    Returns:
        cupy.ndarray: Array of insertion points with the same shape as ``v``.

    .. note:: When a is not in ascending order, behavior is undefined.

    .. seealso:: :func:`numpy.searchsorted`

    """
    return _searchsorted(a, v, side, sorter, True)


def _searchsorted(a, v, side, sorter, assume_increasing):
    """`assume_increasing` is used in the kernel to
    skip monotonically increasing or decreasing verification
    inside the cuda kernel.
    """
    if not isinstance(a, cupy.ndarray):
        raise NotImplementedError('Only int or ndarray are supported for a')

    if not isinstance(v, cupy.ndarray):
        raise NotImplementedError('Only int or ndarray are supported for v')

    if a.ndim > 1:
        raise ValueError('object too deep for desired array')
    if a.ndim < 1:
        raise ValueError('object of too small depth for desired array')
    if a.size == 0:
        return cupy.zeros(v.shape, dtype=cupy.int64)

    a_iscomplex = a.dtype.kind == 'c'
    v_iscomplex = v.dtype.kind == 'c'

    if a_iscomplex and not v_iscomplex:
        v = v.astype(a.dtype)
    elif v_iscomplex and not a_iscomplex:
        a = a.astype(v.dtype)

    # Numpy does not check if the array is monotonic inside searchsorted
    # which leads to undefined behavior in such cases.
    if sorter is not None:
        if sorter.dtype.kind not in ('i', 'u'):
            raise TypeError('sorter must be of integer type')
        if sorter.size != a.size:
            raise ValueError('sorter.size must equal a.size')
        a = a.take(sorter)

    y = cupy.zeros(v.shape, dtype=cupy.int64)

    _searchsorted_kernel(v, a, a.size, side == 'right', assume_increasing, y)
    return y


# TODO(okuta): Implement extract



import numpy
import pytest

import cupy

from cupy import testing
from cupy._core._gufuncs import _GUFunc


class TestGUFuncSignature:
    @pytest.mark.parametrize('signature', [
        ('(i,j)->(i,j)', [('i', 'j')], [('i', 'j')]),
        ('->(i)', [()], [('i',)]),
        ('(i,j),(j,k)->(k,l)', [('i', 'j'), ('j', 'k')], [('k', 'l')]),
        ('()->()', [()], [()])])
    def test_signature_parsing(self, signature):
        i, o = cupy._core._gufuncs._parse_gufunc_signature(signature[0])
        assert i == signature[1]
        assert o == signature[2]

    @pytest.mark.parametrize('signature', [
        '(i,j)(i,j)',
        '(i,j)-(i,j)',
        '(i,j)(i,j)->(i,j)',
        'j->(i',
        '',
        '()->()->'])
    def test_invalid_signature_parsing(self, signature):
        with pytest.raises(ValueError):
            cupy._core._gufuncs._parse_gufunc_signature(signature)


class TestGUFuncAxes:
    def _get_gufunc(self, signature):
        def func(x):
            return x
        return _GUFunc(func, signature)

    def _get_gufunc_scalar(self, signature):
        def func(x):
            return x.sum()
        return _GUFunc(func, signature)

    @pytest.mark.parametrize('axes', [
        ((-1, -2), (-1, -2)),
        ((0, 1), (0, 1)),
        ((0, 1), (-1, -2)),
        ((1, 2), (-1, -2)),
        ((1, 2), (1, 2)),
        ((1, 2), (2, 3)),
        ((2, 3), (-1, -2)),
        ((2, 3), (0, 1)),
        ((2, 3), (1, 2)),
        ((0, 3), (1, 2)),
        ((0, 3), (2, 0)),
    ])
    @testing.numpy_cupy_array_equal()
    def test_axes_selection(self, xp, axes):
        x = testing.shaped_arange((2, 3, 4, 5), xp=xp)
        if xp is cupy:
            return self._get_gufunc('(i,j)->(i,j)')(x, axes=list(axes))
        else:
            return numpy.moveaxis(x, axes[0], axes[1])

    @pytest.mark.parametrize('axes', [
        (-1, -2),
        (0, 1),
        (1, 2),
        (2, 3),
        (0, 2),
        (0, 3),
        (1, 3),
        (3, 0),
        (2, 0),
        (2, 1),
        (1, 0),
    ])
    @testing.numpy_cupy_array_equal()
    def test_axes_selection_single(self, xp, axes):
        x = testing.shaped_arange((2, 3, 4, 5), xp=xp)
        if xp is cupy:
            return self._get_gufunc('(i)->(i)')(x, axes=list(axes))
        else:
            return numpy.moveaxis(x, axes[0], axes[1])

    @pytest.mark.parametrize('axis', [0, 1, 2, 3])
    @testing.numpy_cupy_array_equal()
    def test_axis(self, xp, axis):
        x = testing.shaped_arange((2, 3, 4, 5), xp=xp)
        if xp is cupy:
            return self._get_gufunc_scalar('(i)->()')(x, axis=axis)
        else:
            return x.sum(axis=axis)

    def test_axis_invalid(self):
        x = testing.shaped_arange((2, 3, 4, 5))
        with pytest.raises(ValueError):
            self._get_gufunc('(i, j)->(i, j)')(x, axis=((0, 1), (0, 1)))

    @pytest.mark.parametrize('supports_batched', [True, False])
    def test_supports_batched(self, supports_batched):
        x = testing.shaped_arange((2, 3, 4, 5))

        def func(x):
            nonlocal supports_batched
            if supports_batched:
                assert x.ndim == 4
            else:
                assert x.ndim == 2
            return x
        gu_func = _GUFunc(func, '(i,j)->(i,j)',
                          supports_batched=supports_batched)
        gu_func(x)


class TestGUFuncOut:
    def _get_gufunc(self):
        def func(x):
            return x
        return _GUFunc(func, '(i,j)->(i,j)')

    def test_out_array(self):
        x = testing.shaped_arange((2, 3, 4, 5))
        out = cupy.empty((2, 3, 4, 5))
        self._get_gufunc()(x, out=out)
        testing.assert_allclose(x, out)

    def test_supports_out(self):
        x = testing.shaped_arange((2, 3, 4, 5))
        out = cupy.empty((2, 3, 4, 5))
        out_ptr = out.data.ptr

        def func(x, out=None):
            nonlocal out_ptr
            # Base is a view of the output due to the batching
            assert out.base.data.ptr == out_ptr
            out[:] = x
        gu_func = _GUFunc(func, '(i,j)->(i,j)', supports_out=True)
        gu_func(x, out=out)
        testing.assert_allclose(x, out)

    def test_invalid_output_shape(self):
        x = testing.shaped_arange((2, 3, 4, 5))
        out = cupy.empty((3, 3, 4, 5))
        with pytest.raises(ValueError):
            self._get_gufunc()(x, out=out)

    def test_invalid_output_dtype(self):
        x = testing.shaped_arange((2, 3, 4, 5))
        out = cupy.empty((2, 3, 4, 5), dtype='h')
        with pytest.raises(TypeError):
            self._get_gufunc()(x, out=out)


class TestGUFuncDtype:
    @testing.for_all_dtypes(name='dtype_i', no_bool=True, no_complex=True)
    @testing.for_all_dtypes(name='dtype_o', no_bool=True, no_complex=True)
    def test_dtypes(self, dtype_i, dtype_o):
        x = testing.shaped_arange((2, 3, 4, 5), dtype=dtype_i)
        if numpy.can_cast(dtype_o, x.dtype):
            def func(x):
                return x
            gufunc = _GUFunc(func, '(i,j)->(i,j)')
            z = gufunc(x, dtype=dtype_o)
            assert z.dtype == dtype_o
            testing.assert_allclose(z, x)


class TestGUFuncOrder():

    @pytest.mark.parametrize('order', ['C', 'F', 'K'])
    @testing.numpy_cupy_array_equal(strides_check=True)
    def test_order(self, xp, order):
        x = testing.shaped_arange((2, 3, 4), xp=xp)
        if xp is cupy:
            def default(x):
                return x
            gu_func = _GUFunc(default, '(i, j, k)->(i, j, k)')
            return gu_func(x, order=order)
        else:
            return xp.asarray(x, order=order)

    @pytest.mark.parametrize('order', [('F', 'C', 'C'), ('F', 'F', 'F')])
    def test_order_a(self, order):
        x = testing.shaped_arange((2, 3, 4), order=order[0])
        y = testing.shaped_arange((2, 3, 4), order=order[1])

        def default(x, y):
            return x

        gu_func = _GUFunc(default, '(i,j,k),(i,j,k)->(i,j,k)')
        z = gu_func(x, y, order='A')
        if order[2] == 'C':
            assert z.flags.c_contiguous
        else:
            assert z.flags.f_contiguous


class TestGUFuncSignatures():
    def test_signatures(self):
        dtypes = 'fdihq'
        dtypes_access = {d: None for d in dtypes}

        def integers(x, y):
            nonlocal dtypes_access
            dtypes_access[numpy.dtype(x.dtype).char] = integers
            return x + y

        def floats(x, y):
            nonlocal dtypes_access
            dtypes_access[numpy.dtype(x.dtype).char] = floats
            return x + y

        def default(x, y):
            nonlocal dtypes_access
            dtypes_access[numpy.dtype(x.dtype).char] = default
            return x + y

        sigs = (('ii->i', integers), ('dd->d', floats))
        gu_func = _GUFunc(default, '(i),(i)->(i)', signatures=sigs)
        for dtype in dtypes:
            x = cupy.array([10], dtype=dtype)
            y = x
            gu_func(x, y, casting='no')
            if dtype in 'i':
                assert dtypes_access[dtype] == integers
            elif dtype in 'd':
                assert dtypes_access[dtype] == floats
            else:
                assert dtypes_access[dtype] == default

    @pytest.mark.parametrize('sig,', ['ii->i', 'i', ('i', 'i', 'i')])
    def test_signature_lookup(self, sig):
        called = False

        def func(x, y):
            nonlocal called
            called = True
            return x + y

        def default(x, y):
            return x + y

        dtypes = 'fdhq'

        sigs = (('ii->i', func),)
        gu_func = _GUFunc(default, '(i),(i)->(i)', signatures=sigs)
        for dtype in dtypes:
            x = cupy.array([10], dtype=dtype)
            y = x
            gu_func(x, y, casting='no')
            assert not called

        x = cupy.array([10], dtype='d')
        y = x
        z = gu_func(x, y, casting='unsafe', signature=sig)
        assert z.dtype == numpy.int32
        assert called

    @pytest.mark.parametrize('sigs,', [('i',), ('',), ('iii->i',), ('ii->',)])
    def test_invalid_signatures(self, sigs):

        def default(x, y):
            return x + y

        with pytest.raises(ValueError):
            _GUFunc(default, '(i),(i)->(i)', signatures=sigs)

    @pytest.mark.parametrize('sig,', ['i->i', 'id->i', ''])
    def test_invalid_lookup(self, sig):

        def default(x, y):
            return x + y

        sigs = (('ii->i', default),)
        gu_func = _GUFunc(default, '(i),(i)->(i)', signatures=sigs)
        _GUFunc(default, '(i),(i)->(i)', signatures=sigs)

        x = cupy.array([10], dtype='d')
        y = x
        with pytest.raises(TypeError):
            gu_func(x, y, casting='unsafe', signature=sig)



from cupy import _core


erf = _core.create_ufunc(
    'cupyx_scipy_special_erf', ('f->f', 'd->d'),
    'out0 = erf(in0)',
    doc='''Error function.

    .. seealso:: :meth:`scipy.special.erf`

    ''')


erfc = _core.create_ufunc(
    'cupyx_scipy_special_erfc', ('f->f', 'd->d'),
    'out0 = erfc(in0)',
    doc='''Complementary error function.

    .. seealso:: :meth:`scipy.special.erfc`

    ''')


erfcx = _core.create_ufunc(
    'cupyx_scipy_special_erfcx', ('f->f', 'd->d'),
    'out0 = erfcx(in0)',
    doc='''Scaled complementary error function.

    .. seealso:: :meth:`scipy.special.erfcx`

    ''')


erfinv = _core.create_ufunc(
    'cupyx_scipy_special_erfinv', ('f->f', 'd->d'),
    'out0 = erfinv(in0);',
    doc='''Inverse function of error function.

    .. seealso:: :meth:`scipy.special.erfinv`

    .. note::
        The behavior close to (and outside) the domain follows that of
        SciPy v1.4.0+.

    ''')


erfcinv = _core.create_ufunc(
    'cupyx_scipy_special_erfcinv', ('f->f', 'd->d'),
    'out0 = erfcinv(in0);',
    doc='''Inverse function of complementary error function.

    .. seealso:: :meth:`scipy.special.erfcinv`

    .. note::
        The behavior close to (and outside) the domain follows that of
        SciPy v1.4.0+.

    ''')
